# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module for plotting the data and saving them to disk. This module implements classes and methods for making
smart decisions about plotting various types of plots."""
import numpy as np
import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt
import seaborn as sns
import logging

from learner.configuration.configuration import Configuration
color = sns.color_palette()


class Plot:
    """Implement methods for plotting graphs and saving them to disk."""

    def __init__(self, conf: Configuration):
        """Initialize a Plot object using a conf object.

        :param conf: an instance of the Configuration class
        """
        self._conf = conf

    @property
    def conf(self):
        return self._conf

    def barh(self, x, y, filename, yrange=10, color="blue", width=12, height=6, **kwargs):
        """ Plot a horizontal bar plot using two columns from a pandas dataframe.

        :param x: the x column
        :param y: the y column
        :param filename: the full path to use for saving the plot
        :param yrange: the number of items in the data
        :param color: the color of the bars in the plot
        :param width: the width of the graph
        :param height: the height of the graph
        :param kwargs: otehr kwargs such as xlabel
        :return: None
        """
        try:
            plt.subplots(figsize=(width, height))
            plt.barh(yrange, x.values, color=color)
            plt.yticks(yrange, y.values)
            plt.xlabel(kwargs.get("xlabel", x.name))
            self.savefig(filename, plt)
        except Exception as e:
            logging.error("Unable to plot the results. The error is: {error}".format(error=str(e)))

    def correlation_heatmap(self, data, filename, vmin=None, vmax=None, width=10, height=10, x_labelsize=10, y_labelsize=10, annot_size=10, mask=None):
        """Plot a heatmap using a dataframe. This is currently being used for plotting the correlation coefficient, hence
        the name correlation_heatmap. This can change in the future. This method calls savefig to save the plot as well.

        :param data: a pandas dataframe that contains the data that need to be plotted
        :param filename: the full path to use for saving the plot
        :param vmin: the minimum value to anchor the colormap. If not defined, the value will be inferred internally
        :param vmax: the maximum value to anchor the colormap. If not defined, the value will be inferred internally
        :param width: the width of the graph
        :param height: the height of the graph
        :param x_labelsize: the size of the labels on the x axis
        :param y_labelsize: the size of the labels on the y axis
        :param annot_size: the size of the text that shows the numbers on the graph
        :param mask: the mask matrix to not show data above the diagonal
        :return: None
        """
        try:
            fig = plt.figure(figsize=(width, height))
            fig.patch.set_facecolor('blue')
            plt.rcParams['xtick.labelsize'] = x_labelsize
            plt.rcParams['ytick.labelsize'] = y_labelsize

            with sns.axes_style("white"):
                sns.heatmap(data, vmin=vmin, vmax=vmax, mask=mask, annot=True, annot_kws={"size": annot_size}, cmap='RdBu', fmt='+.2f',
                            cbar=True)
            self.savefig(filename, plt)
        except Exception as e:
            logging.error("Unable to plot the results. The error is: {error}".format(error=str(e)))

    def predictions_vs_actuals(self, x, y, filename, width=10, height=10, x_label="actual", y_label="prediction",
                               x_labelsize=10, y_labelsize=10, color="blue", size=0.1, xy_min=None, xy_max=None):
        """Accept two arrays (x and y) and make a scatter plot as well as a y=x line to show how predictions vs actuals
        look like. After that, save the graph in a file.

        :param x: the x array, this could be a pandas series or a numpy array or a list
        :param y: the x array, this could be a pandas series or a numpy array or a list
        :param filename: the full path for saving the graph
        :param width: the width of the figure
        :param height: the height of the figure
        :param x_label: the label to use for the x axis
        :param y_label: the label to use for the y axis
        :param x_labelsize: the font size for the x label
        :param y_labelsize: the font size for the y label
        :param color: the color of the symbols
        :param size: the size of the symbols
        :return: None
        """
        try:
            fig, ax = plt.subplots(figsize=(width, height))

            ax.set_xlabel(x_label, fontsize=x_labelsize)
            ax.set_ylabel(y_label, fontsize=y_labelsize)

            ax.scatter(x, y, color=color, s=size)

            lims = [
                xy_min or np.min([ax.get_xlim(), ax.get_ylim()]),  # min of both axes
                xy_max or np.max([ax.get_xlim(), ax.get_ylim()]),  # max of both axes
            ]

            ax.plot(lims, lims, 'k-')
            ax.set_xlim(lims)
            ax.set_ylim(lims)

            self.savefig(filename, plt)
        except Exception as e:
            logging.error("Unable to plot the results. The error is: {error}".format(error=str(e)))

    def learning_rate(self, x, y, filename, width=8, height=5, x_label="learning rate", y_label="loss",
                      x_labelsize=10, y_labelsize=10, color="blue"):
        """Accept two arrays (x and y) and make a line plot to show how the loss changes as a function of learning rate.
        Set the x axis to be logarithmic. After that, save the graph in a file.

        :param x: the x array, this could be a pandas series or a numpy array or a list
        :param y: the x array, this could be a pandas series or a numpy array or a list
        :param filename: the full path for saving the graph
        :param width: the width of the figure
        :param height: the height of the figure
        :param x_label: the label to use for the x axis
        :param y_label: the label to use for the y axis
        :param x_labelsize: the font size for the x label
        :param y_labelsize: the font size for the y label
        :param color: the color of the symbols
        :return: None
        """

        try:
            fig, ax = plt.subplots(figsize=(width, height))

            ax.set_xlabel(x_label, fontsize=x_labelsize)
            ax.set_ylabel(y_label, fontsize=y_labelsize)
            ax.set_xscale('log')

            ax.plot(x, y, color=color)
            self.savefig(filename, plt)
        except Exception as e:
            logging.error("Unable to plot the results. The error is: {error}".format(error=str(e)))

    @staticmethod
    def savefig(filename, plt, dpi=600):
        """Given a plt with populated attributes, save the final graph to disk.

        :param filename: the full path to use for saving the plot
        :param dpi: the resolution of the saved image
        :return: None
        """
        try:
            plt.savefig(filename, dpi=dpi, format="png", bbox_inches='tight')
            if hasattr(plt, "close"):
                plt.close()
        except Exception as e:
            logging.error("Unable to save the plot. The error is: {error}".format(error=str(e)))
        logging.info("The plot was successfully saved in {filename}".format(filename=filename))



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import os
import sys
import getpass
import logging
import warnings

import pandas as pd
from sqlalchemy import create_engine

from learner.data_worker.data_loader import get_value
from learner.setup.setup import load_json
from learner.validator.input_validator import InputValidator, validate_subset_list
from learner.configuration.data import DataConfiguration


class ConnectionConfiguration:
    """Parse the fields in the connection section of the configuration file."""

    def __init__(self, json_config, data: DataConfiguration=None):
        self._json_config = json_config
        self._data = data

        self.credentials_schema_file = get_value(self._json_config,
                                                 os.path.join(os.path.dirname(__file__),
                                                              "..", "schema", "credentials_schema.json"),
                                                 "connection",
                                                 "credentials_schema")
        self.credentials_schema = load_json(self.credentials_schema_file)
        self.credentials_file = get_value(self._json_config, os.path.join(os.environ['HOME'], ".learner_credentials.json"),
                                          "connection",
                                          "credentials_file")
        self.credentials = self.get_credentials()
        # this is anti-pattern but looks like the best way to keep things clean
        if self._data is None:  # pragma: no cover
            return
        self.presto_activate = self.get_presto_activate()
        self.presto_protocol = self.get_presto_protocol()
        self.presto_username = self.get_presto_username()
        self.presto_password = self.get_presto_password()
        self.presto_host = self.get_presto_host()
        self.presto_port = get_value(self._json_config, 8080, "connection", "presto_params", "port")
        self.presto_dbname = self.get_presto_dbname()
        self.presto_client = self.get_presto_client()

        self.postgres_activate = self.get_postgres_activate()
        self.postgres_username = self.get_postgres_username()
        self.postgres_password = self.get_postgres_password()
        self.postgres_host = self.get_postgres_host()
        self.postgres_port = get_value(self._json_config, 5432, "connection", "postgres_params", "port")
        self.postgres_dbname = self.get_postgres_dbname()
        self.postgres_client = self.get_postgres_client()

        self.mysql_activate = self.get_mysql_activate()
        self.mysql_username = self.get_mysql_username()
        self.mysql_password = self.get_mysql_password()
        self.mysql_host = self.get_mysql_host()
        self.mysql_port = get_value(self._json_config, 3306, "connection", "mysql_params", "port")
        self.mysql_dbname = self.get_mysql_dbname()
        self.mysql_client = self.get_mysql_client()

        self.snowflake_activate = self.get_snowflake_activate()
        self.snowflake_username = self.get_snowflake_username()
        self.snowflake_password = self.get_snowflake_password()
        self.snowflake_account = self.get_snowflake_account()
        self.snowflake_warehouse = self.get_snowflake_warehouse()
        self.snowflake_role = self.get_snowflake_role()
        self.snowflake_region = self.get_snowflake_region()
        self.snowflake_database = self.get_snowflake_database()
        self.snowflake_schema = self.get_snowflake_schema()
        self.snowflake_client = self.get_snowflake_client()

    def get_credentials(self):
        if self.credentials_file and os.path.exists(self.credentials_file):
            self.credentials = load_json(self.credentials_file)
            logging.info("---Validating the credentials file---")
            InputValidator(self.credentials, self.credentials_schema)
            return self.credentials
        return None  # pragma: no cover

    def get_presto_activate(self):
        activate = False
        try:
            activate = self._json_config["connection"]["presto_params"]["activate"]
        except KeyError:
            activate = False
        finally:
            if not activate:
                if (self._data.train_query_activate and self._data.train_db_type == "presto") or\
                    (self._data.test_query_activate and self._data.test_db_type == "presto"):
                    logging.critical("Connection to a presto database is required based on the information in the "
                                     "data section but presto connection was not activated. Please update the "
                                     "configuration file and try again. Exiting...")
                    sys.exit(1)
            return activate

    def get_presto_protocol(self):
        if self.presto_activate:
            try:
                protocol = self._json_config["connection"]["presto_params"]["protocol"].lower()
                validate_subset_list(parent_list=["http", "https"], parent_name="accepted presto protocols",
                                     subset_list=[protocol], subset_name="provided protocol")
                return protocol
            except KeyError:
                return "http"

    def get_presto_username(self):
        if self.presto_activate and self.presto_protocol == "https":
            username = os.getenv("LEARNER_PRESTO_USERNAME")
            if not username:
                try:
                    username = self._json_config["connection"]["presto_params"]["username"]
                except KeyError:
                    try:
                        username = self.credentials["presto"]["username"]
                    except (KeyError, TypeError):  # pragma: no cover
                        username = input("Please enter your username for presto server: ")
            return username

    def get_presto_password(self):
        if self.presto_activate and self.presto_protocol == "https":
            password = os.getenv("LEARNER_PRESTO_PASSWORD")
            if not password:
                try:
                    password = self._json_config["connection"]["presto_params"]["password"]
                except KeyError:
                    try:
                        password = self.credentials["presto"]["password"]
                    except (KeyError, TypeError):  # pragma: no cover
                        password = getpass.getpass("Please enter your password for presto server: ")
            return password

    def get_presto_host(self):
        if self.presto_activate:
            try:
                return self._json_config["connection"]["presto_params"]["host"]
            except KeyError:
                logging.critical("Connection to a presto server was activated but host is not defined. Please update "
                                 "the configuration file and try again. Exiting...")
                sys.exit(1)

    def get_presto_dbname(self):
        if self.presto_activate:
            try:
                return self._json_config["connection"]["presto_params"]["dbname"]
            except KeyError:
                warnings.warn("Connection to a presto server was activated but dbname is not defined. Leraner will use "
                              "an empty string here", UserWarning)
                return ""

    def get_presto_client(self):
        if self.presto_activate:
            url = self._construct_presto_url()
            client = create_engine(url, connect_args={'protocol': self.presto_protocol})
            self.validate_presto_connection(client)
            return client

    def _construct_presto_url(self):
        if self.presto_protocol == "https":
            url = f"presto://{self.presto_username}:{self.presto_password}@{self.presto_host}:{self.presto_port}/{self.presto_dbname}"
        else:
            url = f"presto://{self.presto_host}:{self.presto_port}/{self.presto_dbname}"
        return url

    @staticmethod
    def validate_presto_connection(client):  # pragma: no cover
        try:
            with client.connect() as conn:
                logging.info("Validating the connection to presto server by running a simple query...")
                result = pd.read_sql("SELECT * FROM system.runtime.nodes LIMIT 1", con=conn)
            assert len(result) > 0, "No running nodes available"
        except Exception as e:
            logging.critical("Test connection to presto server failed. The error is {error}".format(error=str(e)))
            sys.exit(1)

    def get_postgres_activate(self):
        activate = False
        try:
            activate = self._json_config["connection"]["postgres_params"]["activate"]
        except KeyError:
            activate = False
        finally:
            if not activate:
                if (self._data.train_query_activate and self._data.train_db_type == "postgres") or\
                    (self._data.test_query_activate and self._data.test_db_type == "postgres"):
                    logging.critical("Connection to a postgres database is required based on the information in the "
                                     "data section but postgres connection was not activated. Please update the "
                                     "configuration file and try again. Exiting...")
                    sys.exit(1)
            return activate

    def get_postgres_username(self):
        if self.postgres_activate:
            username = os.getenv("LEARNER_POSTGRES_USERNAME")
            if not username:
                try:
                    username = self._json_config["connection"]["postgres_params"]["username"]
                except KeyError:
                    try:
                        username = self.credentials["postgres"]["username"]
                    except (KeyError, TypeError):  # pragma: no cover
                        username = input("Please enter your username for postgres server: ")
            return username

    def get_postgres_password(self):
        if self.postgres_activate:
            password = os.getenv("LEARNER_POSTGRES_PASSWORD")
            if not password:
                try:
                    password = self._json_config["connection"]["postgres_params"]["password"]
                except KeyError:
                    try:
                        password = self.credentials["postgres"]["password"]
                    except (KeyError, TypeError):  # pragma: no cover
                        password = getpass.getpass("Please enter your password for postgres server: ")
            return password

    def get_postgres_host(self):
        if self.postgres_activate:
            try:
                return self._json_config["connection"]["postgres_params"]["host"]
            except KeyError:
                logging.critical("Connection to a postgres server was activated but host is not defined. Please "
                                 "update the configuration file and try again. Exiting...")
                sys.exit(1)

    def get_postgres_dbname(self):
        if self.postgres_activate:
            try:
                return self._json_config["connection"]["postgres_params"]["dbname"]
            except KeyError:
                warnings.warn("Connection to a postgres server was activated but dbname is not defined. Leraner will "
                              "use an empty string here", UserWarning)
                return ""

    def get_postgres_client(self):
        if self.postgres_activate:
            url = self._construct_postgres_url()
            client = create_engine(url)
            self.validate_postgres_connection(client)
            return client

    def _construct_postgres_url(self):
        url = f"postgresql+psycopg2://{self.postgres_username}:{self.postgres_password}@{self.postgres_host}:{self.postgres_port}/{self.postgres_dbname}"
        return url

    @staticmethod
    def validate_postgres_connection(client):  # pragma: no cover
        try:
            with client.connect() as conn:
                logging.info("Validating the connection to postgres server by running a simple query...")
                pd.read_sql("SELECT version()", con=conn)
        except Exception as e:
            logging.critical("Test connection to postgres server failed. The error is {error}".format(error=str(e)))
            sys.exit(1)

    def get_mysql_activate(self):
        activate = False
        try:
            activate = self._json_config["connection"]["mysql_params"]["activate"]
        except KeyError:
            activate = False
        finally:
            if not activate:
                if (self._data.train_query_activate and self._data.train_db_type == "mysql") or \
                        (self._data.test_query_activate and self._data.test_db_type == "mysql"):
                    logging.critical("Connection to a mysql database is required based on the information in the data "
                                     "section but mysql connection was not activated. Please update the configuration "
                                     "file and try again. Exiting...")
                    sys.exit(1)
            return activate

    def get_mysql_username(self):
        if self.mysql_activate:
            username = os.getenv("LEARNER_MYSQL_USERNAME")
            if not username:
                try:
                    username = self._json_config["connection"]["mysql_params"]["username"]
                except KeyError:
                    try:
                        username = self.credentials["mysql"]["username"]
                    except (KeyError, TypeError):  # pragma: no cover
                        username = input("Please enter your username for mysql server: ")
            return username

    def get_mysql_password(self):
        if self.mysql_activate:
            password = os.getenv("LEARNER_MYSQL_PASSWORD")
            if not password:
                try:
                    password = self._json_config["connection"]["mysql_params"]["password"]
                except KeyError:
                    try:
                        password = self.credentials["mysql"]["password"]
                    except (KeyError, TypeError):  # pragma: no cover
                        password = getpass.getpass("Please enter your password for mysql server: ")
            return password

    def get_mysql_host(self):
        if self.mysql_activate:
            try:
                return self._json_config["connection"]["mysql_params"]["host"]
            except KeyError:
                logging.critical("Connection to a mysql server was activated but host is not defined. Please update "
                                 "the configuration file and try again. Exiting...")
                sys.exit(1)

    def get_mysql_dbname(self):
        if self.mysql_activate:
            try:
                return self._json_config["connection"]["mysql_params"]["dbname"]
            except KeyError:
                warnings.warn("Connection to a mysql server was activated but dbname is not defined. Leraner will "
                              "use an empty string here", UserWarning)
                return ""

    def get_mysql_client(self):
        if self.mysql_activate:
            url = self._construct_mysql_url()
            client = create_engine(url)
            self.validate_mysql_connection(client)
            return client

    def _construct_mysql_url(self):
        url = f"mysql+mysqlconnector://{self.mysql_username}:{self.mysql_password}@{self.mysql_host}:{self.mysql_port}/{self.mysql_dbname}"
        return url

    @staticmethod
    def validate_mysql_connection(client):  # pragma: no cover
        try:
            with client.connect() as conn:
                logging.info("Validating the connection to mysql server by running a simple query...")
                pd.read_sql("SELECT version()", con=conn)
        except Exception as e:
            logging.critical("Test connection to mysql server failed. The error is {error}".format(error=str(e)))
            sys.exit(1)

    def get_snowflake_activate(self):
        activate = False
        try:
            activate = self._json_config["connection"]["snowflake_params"]["activate"]
        except KeyError:
            activate = False
        finally:
            if not activate:
                if (self._data.train_query_activate and self._data.train_db_type == "snowflake") or \
                        (self._data.test_query_activate and self._data.test_db_type == "snowflake"):
                    logging.critical("Connection to a snowflake database is required based on the information in the "
                                     "data section but snowflake connection was not activated. Please update the "
                                     "configuration file and try again. Exiting...")
                    sys.exit(1)
            return activate

    def get_snowflake_username(self):
        if self.snowflake_activate:
            username = os.getenv("LEARNER_SNOWFLAKE_USERNAME")
            if not username:
                try:
                    username = self._json_config["connection"]["snowflake_params"]["username"]
                except KeyError:
                    try:
                        username = self.credentials["snowflake"]["username"]
                    except (KeyError, TypeError):  # pragma: no cover
                        username = input("Please enter your username for snowflake server: ")
            return username

    def get_snowflake_password(self):
        if self.snowflake_activate:
            password = os.getenv("LEARNER_SNOWFLAKE_PASSWORD")
            if not password:
                try:
                    password = self._json_config["connection"]["snowflake_params"]["password"]
                except KeyError:
                    try:
                        password = self.credentials["snowflake"]["password"]
                    except (KeyError, TypeError):  # pragma: no cover
                        password = getpass.getpass("Please enter your password for snowflake server: ")
            return password

    def get_snowflake_warehouse(self):
        if self.snowflake_activate:
            try:
                return self._json_config["connection"]["snowflake_params"]["warehouse"]
            except KeyError:
                return None

    def get_snowflake_role(self):
        if self.snowflake_activate:
            try:
                return self._json_config["connection"]["snowflake_params"]["role"]
            except KeyError:
                return None

    def get_snowflake_database(self):
        if self.snowflake_activate:
            try:
                return self._json_config["connection"]["snowflake_params"]["database"]
            except KeyError:
                return None

    def get_snowflake_schema(self):
        if self.snowflake_activate:
            try:
                return self._json_config["connection"]["snowflake_params"]["schema"]
            except KeyError:
                return None

    def get_snowflake_region(self):
        if self.snowflake_activate:
            try:
                return self._json_config["connection"]["snowflake_params"]["region"]
            except KeyError:
                return None

    def get_snowflake_account(self):
        if self.snowflake_activate:
            try:
                return self._json_config["connection"]["snowflake_params"]["account"]
            except KeyError:
                logging.critical("Connection to a snowflake server was activated but account is not defined. "
                                 "Please update the configuration file and try again. Exiting...")
                sys.exit(1)

    def get_snowflake_client(self):
        if self.snowflake_activate:
            url = self._construct_snowflake_url()
            client = create_engine(url)
            self.validate_snowflake_connection(client)
            return client

    def _construct_snowflake_url(self):
        from snowflake.sqlalchemy import URL
        params = {"account": self.snowflake_account,
                  "user": self.snowflake_username,
                  "password": self.snowflake_password}
        if self.snowflake_account:
            params["account"] = self.snowflake_account
        if self.snowflake_region:
            params["region"] = self.snowflake_region
        if self.snowflake_database:
            params["database"] = self.snowflake_database
        if self.snowflake_schema:
            if not self.snowflake_database:
                logging.critical("snowflake schema was defined but database was not. Please update your configuration "
                                 "file and try again. Exiting...")
                sys.exit(1)
            params["schema"] = self.snowflake_schema
        if self.snowflake_role:
            params["role"] = self.snowflake_role

        return URL(**params)

    @staticmethod
    def validate_snowflake_connection(client):  # pragma: no cover
        try:
            with client.connect() as conn:
                logging.info("Validating the connection to snowflake server by running a simple query...")
                pd.read_sql("SELECT current_version()", con=conn)
        except Exception as e:
            logging.critical("Test connection to snowflake server failed. The error is {error}".format(error=str(e)))
            sys.exit(1)



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This class implements the DeepTrainManager class to train and validate deep learning models."""
import logging
import numpy as np
import torch
from torch.nn.functional import softmax


from learner.callback_manager.callback import MetricLogger, CallbackManager, ModelSaver
from learner.model_manager.scoring_manager import DeepClassifierScorer, DeepRegressorScorer
from learner.data_worker.data_mover import move_to_device


class DeepLoopManager:
    """The DeepTrainManager implements the main loops for building deep learning models. This class contains two main
    methods namely "train" and "validation". This class uses a callback/event handling process to modify the
    training/validation loops, log the outputs, save the models or any other operations. Currently, the "validate"
    method is specific to the image_classifier engine but this will be updated once we implement other deep learning
    engines.
    """
    def __init__(self, conf, model, criterion, optimizer, scheduler, epochs, tag, mdl, train_loader=None,
                 validation_loader=None, callbacks=None, processor=None, feature_engineering=None):
        """Initialize a DeepTrainManager using the input arguments.

        :param conf: a conf object
        :param model: a PyTorch model that is ready to be trained.
        :param criterion: the loss object. This is built using the loss_manager module
        :param optimizer: the optimizer object. This is built using the optimizer_manager module.
        :param epochs: the number of epochs to train the model for. This is an integer.
        :param tag: the tag of the model. This is a key in models_dict.
        :param mdl: an item (value) in models_dict.
        :param train_loader: the train_loader to loop through the training data.
        :param validation_loader: the validation_loader to loop through the validation data.
        :param callbacks: a list of callbacks
        :param processor: the processor object - this is optional and only relevant for deep_classifier and deep_regressor engines
        :param feature_engineering: the feature_engineering object - this is optional and only relevant for deep_classifier and deep_regressor engines
        """
        self.conf = conf
        self.epochs = epochs
        self.tag = tag
        self.mdl = mdl
        self.train_loader = train_loader
        self.validation_loader = validation_loader
        self.device = conf.model.device
        self.model = model.to(self.device)
        self.criterion = criterion
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.callbacks = callbacks
        self.processor = processor
        self.feature_engineering = feature_engineering
        self.metrics = {}
        # the Event classes can modify this to stop the process when needed
        self.should_stop = False

    def train(self):
        """Implement the main training loop. Here, we first use a list of event objects to create an event_manager
        object. Currently, we perform this operation in this methods but we may need to change this in the future.
        We then perform the main training loop and call the "validation" method after each epoch.

        :return: None
        """
        event_manager = CallbackManager(*self.callbacks)
        logging.info(f"Training {self.tag} model ")
        for epoch in range(self.epochs):
            self.metrics["epoch"] = epoch
            self.model.train()

            train_losses = np.zeros(len(self.train_loader))
            for b, (X_train, y_train) in enumerate(self.train_loader):
                X_train, y_train = move_to_device(X_train, self.device), move_to_device(y_train, self.device)

                y_pred = self.model(*X_train) if isinstance(X_train, list) else self.model(X_train)

                loss = self.criterion(y_pred, y_train)
                train_losses[b] = loss

                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

                event_manager.on_batch_end(deep_trainer=self,
                                           iteration=epoch*len(self.train_loader)+b,
                                           loss=loss)
                if self.should_stop: return

            self.metrics["train_loss"] = train_losses.mean()
            self.validate()
            if self.scheduler: self.scheduler.step()

            event_manager.on_epoch_end(conf=self.conf,
                                       tag=self.tag,
                                       mdl=self.mdl,
                                       metrics=self.metrics,
                                       model=self.model,
                                       epoch=epoch,
                                       processor=self.processor,
                                       feature_engineering=self.feature_engineering)
            if self.should_stop: return

    def validate(self):
        """If validation_loader is provided, we need to put the model in validation mode to validate it against the
        validation data. Since there are some nuances in the classification vs regression models, we implemented
        specific methods for those models. In the future, we may consider implementing subclasses for this.

        :return: None
        """
        # return if we don't have validation loader, we can't score
        if not self.validation_loader:
            return

        if self.conf.engine in ["DeepClassifier", "ImageClassifier"]:
            self._validate_classifier()
        else:
            self._validate_regressor()

    def _validate_classifier(self):
        """We first put the model in validation mode to validate it against the validation data. We then run the
        validation loop to get the validation data and compute the scores. This method is specifically designed for the
        classification as it deals with the probabilities and uses the classification scorer class.

        :return: None
        """
        y_true = []
        pred_proba = []
        with torch.no_grad():
            # put the model to evaluation mode. This is very important otherwise the results will not be correct
            self.model.eval()
            validation_losses = np.zeros(len(self.train_loader))
            for b, (X_test, y_test) in enumerate(self.train_loader):
                X_test, y_test = move_to_device(X_test, self.device), move_to_device(y_test, self.device)

                y_val = self.model(*X_test) if isinstance(X_test, list) else self.model(X_test)

                loss = self.criterion(y_val, y_test)

                y_val = self._handle_engine_y_val(y_val)
                y_true.extend(y_test.cpu().numpy())
                pred_proba.extend(y_val.cpu().numpy())

                validation_losses[b] = loss

            self.metrics["validation_loss"] = validation_losses.mean()
            scorer = DeepClassifierScorer(conf=self.conf,
                                          models_dict=None,
                                          y_true=y_true,
                                          pred_proba=pred_proba,
                                          metrics=self.metrics)
            scorer.score()

    def _validate_regressor(self):
        """We first put the model in validation mode to validate it against the validation data. We then run the
        validation loop to get the validation data and compute the scores. This method is specifically designed for the
        regression and uses the regression scorer class.

        :return: None
        """
        y_true = []
        pred = []
        with torch.no_grad():
            # put the model to evaluation mode. This is very important otherwise the results will not be correct
            self.model.eval()
            validation_losses = np.zeros(len(self.train_loader))
            for b, (X_test, y_test) in enumerate(self.train_loader):
                X_test, y_test = move_to_device(X_test, self.device), move_to_device(y_test, self.device)

                y_val = self.model(*X_test) if isinstance(X_test, list) else self.model(X_test)

                loss = self.criterion(y_val, y_test)

                y_true.extend(y_test.cpu().numpy())
                pred.extend(y_val.cpu().numpy())

                validation_losses[b] = loss

            self.metrics["validation_loss"] = validation_losses.mean()
            scorer = DeepRegressorScorer(conf=self.conf,
                                         models_dict=None,
                                         y_true=y_true,
                                         pred=pred,
                                         metrics=self.metrics)
            scorer.score()

    def _handle_engine_y_val(self, y_val):
        """Depending on the type of the models and their last layer, we may need to process the output if the last
        layer in order to correctly compute the scores. For example, if the last layer is LogSoftmax, we need to do the
        exp to get the probabilities. This method handles those calculations. Currently, the regression models don't
        need any additional calculations.

        :param y_val:
        :return:
        """
        if self.conf.engine == "ImageClassifier":
            # if the activation is LogSoftmax, do the exponential to get the probabilities
            from learner.model_manager.image_classifiers import mdl_to_head
            if isinstance(getattr(self.model, mdl_to_head[self.mdl["type"]])[-1], torch.nn.LogSoftmax):
                y_val = torch.exp(y_val)
            # if activation is not LogSoftmax and not Softmax, we need to pass it through a softmax function to get
            # the probabilities
            if not isinstance(getattr(self.model, mdl_to_head[self.mdl["type"]])[-1], torch.nn.LogSoftmax) and \
                    not isinstance(getattr(self.model, mdl_to_head[self.mdl["type"]])[-1], torch.nn.Softmax):
                y_val = torch.nn.functional.softmax(y_val, dim=1)
        elif self.conf.engine == "DeepClassifier":
            # if the activation is LogSoftmax, do the exponential to get the probabilities
            if isinstance(getattr(self.model, "fully_connected_layers")[-1], torch.nn.LogSoftmax):
                y_val = torch.exp(y_val)
            # if activation is not LogSoftmax and not Softmax, we need to pass it through a softmax function to get
            # the probabilities
            if not isinstance(getattr(self.model, "fully_connected_layers")[-1], torch.nn.LogSoftmax) and \
                    not isinstance(getattr(self.model, "fully_connected_layers")[-1], torch.nn.Softmax):
                y_val = torch.nn.functional.softmax(y_val, dim=1)
        return y_val



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import sys
from math import e
import warnings
import logging

from learner.validator.input_validator import validate_subset_list
from learner.configuration.supported_items import SUPPORTED_GROUPBY_AGGREGATION_FUNCTIONS


class FeatureEngineeringConfiguration:
    """Parse all inputs related to feature_engineering section of the configuration file."""

    def __init__(self, json_config, column, process):
        self._column = column
        self._process = process
        self._json_config = json_config
        self.supported_basic_operations = {"m": "multiplication",
                                           "d": "division",
                                           "a": "addition",
                                           "s": "subtraction"}
        self.drop_cols = []
        self.col_dict = {}
        self.fillnan_mean_cols = []
        self.fillnan_median_cols = []
        self.fillnan_mode_cols = []
        self.fillnan_value_cols = {}
        self.basic_operations_params = self.get_basic_operations_params()

        self.log_transform_params = self.get_log_transform_params()
        self.exponential_transform_params = self.get_exponential_transform_params()
        self.groupby_params = self.get_groupby_params()

    def get_basic_operations_params(self):
        try:
            basic_operations_params = self._json_config["feature_engineering"]["basic_operations_params"]
            self.remove_inactive_items(basic_operations_params)
            self.update_method_names(basic_operations_params)
            self.validate_basic_operations_cols_length(basic_operations_params)
            names = self.get_update_basic_operations_col_names(basic_operations_params)
            self.update_col_dicts(names)
            self.update_drop_cols_using_basic_operations(basic_operations_params)
            self._column.valid_cols.extend(names)

            return basic_operations_params
        except KeyError:
            return None

    @staticmethod
    def remove_inactive_items(params):
        i = 0
        while 0 <= i < len(params):
            try:
                if params[i]["activate"] is False:
                    del params[i]
                    i -= 1
                i += 1
            except KeyError:
                del params[i]
                i -= 1

    def update_method_names(self, params):
        for item in params:
            try:
                operation = item["method"].lower()
                if operation not in self.supported_basic_operations.values():
                    try:
                        item["method"] = self.supported_basic_operations[operation[0]]
                    except KeyError:
                        logging.error('The basic_operation method "%s" not understood, supported operations are %s',
                                      operation, list(self.supported_basic_operations.values()))
                        sys.exit(1)
                item["method"] = self.supported_basic_operations[operation[0]]
            except (KeyError, IndexError):
                logging.error("The method name in basic_operations must be defined, the supported methods are %s. "
                              "Exiting...", list(self.supported_basic_operations.values()))
                sys.exit(1)

    @staticmethod
    def validate_basic_operations_cols_length(params):
        for item in params:
            try:
                length = len(item["cols"])
                first_col = item["cols"][0]
                if length == 1 and "value" not in item:
                    item["cols"].append(first_col)
                    warnings.warn("Only one column was defined without any constant value, "
                                  "duplicating the same column for basic operation",
                                  UserWarning)
                elif length > 1 and "value" in item:
                    logging.critical("Only one column should be defined when 'value' exist in the basic_operations "
                                     "list. Please update your configuration file and try again. Exiting...")
                    sys.exit(1)
            except (KeyError, IndexError):
                logging.critical("At least one column must be defined for each item in basic_operations_params. "
                              "Nothing was found. Exiting...")
                sys.exit(1)

    @staticmethod
    def get_update_basic_operations_col_names(params):
        names = []
        for item in params:
            try:
                name = item["name"]
                names.append(name)
            except KeyError:
                method = item["method"]
                # if we get here and only have one item in "cols" that mean we also have "value" key as well.
                # We use it to construct the name
                if len(item["cols"]) == 1 and "value" in item:
                    name = item["cols"][0] + "__" + method[0] + "__" + str(item["value"])
                else:
                    name = ''
                    for i, col in enumerate(item["cols"], 1):
                        if i < len(item["cols"]):
                            name += col + "__" + method[0] + '__'
                        else:
                            name += col
                names.append(name)
                item["name"] = name
        return names

    def update_col_dicts(self, names):
        for name in names:
            if self._process.fillnan_mean_cols and name in self._process.fillnan_mean_cols:
                self.col_dict[name] = "mean"
                self.fillnan_mean_cols.append(name)
                self._process.fillnan_mean_cols.remove(name)
            if self._process.fillnan_median_cols and name in self._process.fillnan_median_cols:
                self.col_dict[name] = "median"
                self.fillnan_median_cols.append(name)
                self._process.fillnan_median_cols.remove(name)
            if self._process.fillnan_mode_cols and name in self._process.fillnan_mode_cols:
                self.col_dict[name] = "mode"
                self.fillnan_mode_cols.append(name)
                self._process.fillnan_mode_cols.remove(name)
            if self._process.fillnan_value_cols and name in self._process.fillnan_value_cols:
                self.col_dict[name] = self._process.fillnan_value_cols[name]
                self.fillnan_value_cols[name] = self._process.fillnan_value_cols[name]
                del self._process.fillnan_value_cols[name]

    def update_drop_cols_using_basic_operations(self, params):
        self.drop_cols = set(self.drop_cols)
        for item in params:
            try:
                if item["drop"]:
                    self.drop_cols = self.drop_cols.union(set(item["cols"]))
                else:
                    warnings.warn("The cols {0} will not be dropped after performing feature engineering".format(
                        item["cols"]), UserWarning)
            except KeyError:
                warnings.warn("The cols {0} will not be dropped after performing feature engineering".format(
                    item["cols"]), UserWarning)

        self.drop_cols = list(self.drop_cols)

    def get_log_transform_params(self):
        try:
            log_transform_params = self._json_config["feature_engineering"]["log_transform_params"]
            self.remove_inactive_items(log_transform_params)
            self.validate_log_transform_cols(log_transform_params)
            self.update_base_values(log_transform_params)
            self.update_shift_values(log_transform_params)
            names = self.get_update_log_transform_col_names(log_transform_params)
            self.update_col_dicts(names)
            self.update_drop_cols_using_log_transform(log_transform_params)
            self._column.valid_cols.extend(names)

            return log_transform_params
        except KeyError:
            return None

    @staticmethod
    def get_update_log_transform_col_names(params):
        names = []
        for item in params:
            try:
                name = item["name"]
                names.append(name)
            except KeyError:
                name = "log_" + item["col"]
                names.append(name)
                item["name"] = name
        return names

    @staticmethod
    def validate_log_transform_cols(params):
        for item in params:
            if "col" not in item:
                logging.critical(f"'col' must be defined in all active items in log_transform_params. Please check  "
                                 f"{item} and try again. Exiting...")
                sys.exit(1)

    @staticmethod
    def update_base_values(params):
        for item in params:
            if "base" not in item:
                item["base"] = e

    @staticmethod
    def update_shift_values(params):
        for item in params:
            if "shift" not in item:
                item["shift"] = 0

    def update_drop_cols_using_log_transform(self, params):
        self.drop_cols = set(self.drop_cols)
        for item in params:
            try:
                if item["drop"]:
                    self.drop_cols = self.drop_cols.union([item["col"]])
                else:
                    warnings.warn("The cols {0} will not be dropped after performing feature engineering".format(
                        item["col"]), UserWarning)
            except KeyError:
                warnings.warn("The cols {0} will not be dropped after performing feature engineering".format(
                    item["col"]), UserWarning)

        self.drop_cols = list(self.drop_cols)

    def get_exponential_transform_params(self):
        try:
            exponential_transform_params = self._json_config["feature_engineering"]["exponential_transform_params"]
            self.remove_inactive_items(exponential_transform_params)
            self.validate_exponential_transform_cols(exponential_transform_params)
            self.update_power_values(exponential_transform_params)
            self.update_shift_values(exponential_transform_params)
            names = self.get_update_exponential_transform_col_names(exponential_transform_params)
            self.update_col_dicts(names)
            self.update_drop_cols_using_exponential_transform(exponential_transform_params)
            self._column.valid_cols.extend(names)

            return exponential_transform_params
        except KeyError:
            return None

    @staticmethod
    def get_update_exponential_transform_col_names(params):
        names = []
        for item in params:
            try:
                name = item["name"]
                names.append(name)
            except KeyError:
                name = "exponential_" + item["col"]
                names.append(name)
                item["name"] = name
        return names

    @staticmethod
    def validate_exponential_transform_cols(params):
        for item in params:
            if "col" not in item:
                logging.critical(f"'col' must be defined in all active items in exponential_transform_params. Please check  "
                                 f"{item} and try again. Exiting...")
                sys.exit(1)

    @staticmethod
    def update_power_values(params):
        for item in params:
            if "power" not in item:
                item["power"] = e

    def update_drop_cols_using_exponential_transform(self, params):
        self.drop_cols = set(self.drop_cols)
        for item in params:
            try:
                if item["drop"]:
                    self.drop_cols = self.drop_cols.union([item["col"]])
                else:
                    warnings.warn("The cols {0} will not be dropped after performing feature engineering".format(
                        item["col"]), UserWarning)
            except KeyError:
                warnings.warn("The cols {0} will not be dropped after performing feature engineering".format(
                    item["col"]), UserWarning)

        self.drop_cols = list(self.drop_cols)

    def get_groupby_params(self):
        try:
            groupby_params = self._json_config["feature_engineering"]["groupby_params"]
            self.remove_inactive_items(groupby_params)
            self.validate_groupby_cols(groupby_params)
            self.validate_groupby_method_names(groupby_params)
            names = self.get_update_groupby_col_names(groupby_params)
            self.update_col_dicts(names)
            self._column.valid_cols.extend(names)
            return groupby_params
        except KeyError:
            return None

    @staticmethod
    def validate_groupby_cols(params):
        for param in params:
            if "col" not in param:
                logging.critical(f"'col' must be defined in all active items in groupby_params. Please check  "
                                 f"{param} and try again. Exiting...")
                sys.exit(1)
            if "aggregation" not in param:
                logging.critical(f"'aggregation' list must be defined in all active items in groupby_params. Please check  "
                                 f"{param} and try again. Exiting...")
                sys.exit(1)
            for agg in param["aggregation"]:
                if "col" not in agg:
                    logging.critical(f"'col' must be defined in each item in 'aggregation' list. Please check  "
                                     f"{param} and try again. Exiting...")
                    sys.exit(1)

    @staticmethod
    def validate_groupby_method_names(params):
        for param in params:
            for agg in param["aggregation"]:
                try:
                    method = agg["method"]
                    validate_subset_list(parent_list=SUPPORTED_GROUPBY_AGGREGATION_FUNCTIONS,
                                         parent_name="supported aggregation methods",
                                         subset_list=[method],
                                         subset_name="defined method")
                except KeyError:
                    logging.error("The method name must be defined for each item in 'aggregation' list. "
                                  "The supported methods are {supported_methods}. Please update your configuration "
                                  "file and try again. Exiting...".format(supported_methods=SUPPORTED_GROUPBY_AGGREGATION_FUNCTIONS))
                    sys.exit(1)

    @staticmethod
    def get_update_groupby_col_names(params):
        names = []
        for param in params:
            for agg in param["aggregation"]:
                try:
                    name = agg["name"]
                    names.append(name)
                except KeyError:
                    name = f"{agg['method']}_{agg['col']}_groupby_{param['col']}"
                    names.append(name)
                    agg["name"] = name
        return names



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import sys
import os
import re
import warnings
import logging

from learner.data_worker.data_loader import get_value
from learner.configuration.workspace import WorkspaceConfiguration
from learner.configuration.data import DataConfiguration
from learner.configuration.column import ColumnConfiguration
from learner.validator.input_validator import validate_subset_list, validate_intersection_cols
from learner.configuration.supported_items import SUPPORTED_NARRATIVE_METHODS, SUPPORTED_ENGINE_FOR_PREDICTIONS_VS_ACTUALS_PLOT


class AnalysisConfiguration:
    def __init__(self, json_config, column: ColumnConfiguration, workspace: WorkspaceConfiguration,
                 data: DataConfiguration, sep_timetag, engine):
        self._json_config = json_config
        self._column = column
        self._workspace = workspace
        self._data = data
        self._sep_timetag = sep_timetag
        self._engine = engine

        self.importance_activate = get_value(self._json_config, False, "analysis", "feature_importance_params", "activate")
        self.importance_plot_activate = get_value(self._json_config, False, "analysis", "feature_importance_params", "plot_params", "activate")
        self.importance_plot_num_features = get_value(self._json_config, 10, "analysis", "feature_importance_params", "plot_params", "num_features")
        self.importance_plot_color = get_value(self._json_config, "blue", "analysis", "feature_importance_params", "plot_params", "options", "color")
        self.importance_plot_width = get_value(self._json_config, None, "analysis", "feature_importance_params", "plot_params", "options", "width")
        self.importance_plot_height = get_value(self._json_config, None, "analysis", "feature_importance_params", "plot_params", "options", "height")

        self.analysis_directory = self.get_analysis_directory()

        self.missing_activate = get_value(self._json_config, False, "analysis", "missing_values_params", "activate")
        self.missing_plot_activate = get_value(self._json_config, False, "analysis", "missing_values_params", "plot_params", "activate")
        self.missing_plot_num_features = get_value(self._json_config, 10, "analysis", "missing_values_params", "plot_params", "num_features")
        self.missing_plot_color = get_value(self._json_config, "blue", "analysis", "missing_values_params", "plot_params", "options", "color")
        self.missing_plot_width = get_value(self._json_config, None, "analysis", "missing_values_params", "plot_params", "options", "width")
        self.missing_plot_height = get_value(self._json_config, None, "analysis", "missing_values_params", "plot_params", "options", "height")
        self.missing_filename = "missing_values{sep_timetag}.csv".format(sep_timetag=self.sep_timetag)
        self.missing_plotname = "missing_values{sep_timetag}.png".format(sep_timetag=self.sep_timetag)

        self.correlation_activate = get_value(self._json_config, False, "analysis", "correlation_params", "activate")
        self.correlation_plot_activate = get_value(self._json_config, False, "analysis", "correlation_params", "plot_params", "activate")
        self.correlation_cols = self.get_correlation_cols()
        self.correlation_plot_width = get_value(self._json_config, None, "analysis", "correlation_params", "plot_params", "options", "width")
        self.correlation_plot_height = get_value(self._json_config, None, "analysis", "correlation_params", "plot_params", "options", "height")
        self.correlation_plot_vmin = get_value(self._json_config, None, "analysis", "correlation_params", "plot_params", "options", "vmin")
        self.correlation_plot_vmax = get_value(self._json_config, None, "analysis", "correlation_params", "plot_params", "options", "vmax")
        self.correlation_plot_x_labelsize = get_value(self._json_config, None, "analysis", "correlation_params", "plot_params", "options", "x_labelsize")
        self.correlation_plot_y_labelsize = get_value(self._json_config, None, "analysis", "correlation_params", "plot_params", "options", "y_labelsize")
        self.correlation_plot_annotation_size = get_value(self._json_config, None, "analysis", "correlation_params", "plot_params", "options", "annotation_size")
        self.correlation_filename = "correlation{sep_timetag}.csv".format(sep_timetag=self.sep_timetag)
        self.correlation_plotname = "correlation{sep_timetag}.png".format(sep_timetag=self.sep_timetag)

        self.shap_activate = self.get_shap_activate()
        self.shap_num_features = get_value(self._json_config, 10, "analysis", "shap_params", "num_features")
        self.shap_narrative_activate = get_value(self._json_config, False, "analysis", "shap_params", "narrative_params", "activate")
        self.shap_include_raw_values = get_value(self._json_config, False, "analysis", "shap_params", "include_raw_values")
        self.shap_plot_activate = get_value(self._json_config, False, "analysis", "shap_params", "plot_params", "activate")
        self.shap_include_sorted_values = self.get_shap_include_sorted_values()
        self.shap_exclude_cols = self.get_shap_exclude_cols()
        self.shap_use_training_mean = get_value(self._json_config, False, "analysis", "shap_params", "use_training_mean")
        self.shap_narrative_stories = self.get_shape_narrative_stories()
        self.shap_narrative_method = self.get_shape_narrative_method()

        self.predictions_vs_actuals_activate = self.get_predictions_vs_actuals_activate()
        self.predictions_vs_actuals_width = get_value(self._json_config, 10, "analysis", "predictions_vs_actuals_params", "width")
        self.predictions_vs_actuals_height = get_value(self._json_config, 10, "analysis", "predictions_vs_actuals_params", "height")
        self.predictions_vs_actuals_x_labelsize = get_value(self._json_config, 10, "analysis", "predictions_vs_actuals_params", "x_labelsize")
        self.predictions_vs_actuals_y_labelsize = get_value(self._json_config, 10, "analysis", "predictions_vs_actuals_params", "y_labelsize")
        self.predictions_vs_actuals_x_label = get_value(self._json_config, None, "analysis", "predictions_vs_actuals_params", "x_label")
        self.predictions_vs_actuals_y_label = get_value(self._json_config, None, "analysis", "predictions_vs_actuals_params", "y_label")
        self.predictions_vs_actuals_xy_min = get_value(self._json_config, None, "analysis", "predictions_vs_actuals_params", "xy_min")
        self.predictions_vs_actuals_xy_max = get_value(self._json_config, None, "analysis", "predictions_vs_actuals_params", "xy_max")
        self.predictions_vs_actuals_symbol_size = get_value(self._json_config, 1, "analysis", "predictions_vs_actuals_params", "symbol_size")
        self.predictions_vs_actuals_symbol_color = get_value(self._json_config, "blue", "analysis", "predictions_vs_actuals_params", "symbol_color")

    @property
    def json_config(self):  # pragma: no cover
        return self._json_config

    @property
    def output(self):
        return self._workspace

    @property
    def sep_timetag(self):
        return self._sep_timetag

    def get_analysis_directory(self):  # pragma: no cover
        # create a directory called "analysis" under output.path to save the analysis data and plots. Note that feature
        # importance does not stay in this directory because it is specific to each model.
        analysis_directory = self.output.path + "analysis/"
        if not os.path.exists(analysis_directory):
            os.makedirs(analysis_directory)
        return analysis_directory

    def get_correlation_cols(self):
        try:
            correlation_cols = self._json_config["analysis"]["correlation_params"]["cols"]
            return correlation_cols
        except KeyError:
            return self._column.use_cols

    def get_shap_activate(self):
        try:
            activate = self._json_config["analysis"]["shap_params"]["activate"]
            if activate:
                warnings.warn("shap analysis was activated. Learner will attempt to load the entire test dataset. "
                              "Please ensure you have enough resources available", Warning)
                # setting the chunksize to a big number so that the entire test data will be loaded
                self._data.test_chunksize = 9000000000
            return activate
        except KeyError:
            return False

    def get_shap_include_sorted_values(self):
        include_sorted_values = get_value(self._json_config, False, "analysis", "shap_params", "include_sorted_values")
        if self.shap_activate and not (include_sorted_values or self.shap_include_raw_values or
                                       self.shap_narrative_activate or self.shap_plot_activate):
            logging.critical("shape feature was activated but include_sorted_values and include_raw_values were "
                             "set to false and narrative was not activated. Please set one of these to true or "
                             "deactivate shap feature. Exiting...")
            sys.exit(1)
        return include_sorted_values

    def get_shap_exclude_cols(self):
        try:
            exclude_cols = self._json_config["analysis"]["shap_params"]["exclude_cols"]
            return exclude_cols
        except KeyError:
            return []

    def get_shape_narrative_stories(self):
        if not self.shap_narrative_activate:
            return None
        # this is the pattern for finding the values between curly braces
        pattern = re.compile("{(.*?)}")
        try:
            shap_narrative_stories = self._json_config["analysis"]["shap_params"]["narrative_params"]["stories"]
            # if there's only one item, make sure the key is called "target"
            if len(shap_narrative_stories) == 1:
                shap_narrative_stories["target"] = shap_narrative_stories.pop(next(iter(shap_narrative_stories)))
            # make sure each item has "templates" and "generic" keys
            for class_name, stories in shap_narrative_stories.items():
                validate_subset_list(parent_list=["templates", "generic"], parent_name="required items in stories",
                                     subset_list=stories.keys(), subset_name="provided items")
                # these two lines make generic and template keys optional
                stories["generic"] = stories.get("generic", None)
                stories["templates"] = stories.get("templates", [])
                # make sure there's no template in the generic text
                results = re.findall(pattern, stories["generic"]) if stories["generic"] else None
                if results:
                    logging.error("Learner does not accept template in generic text. These items were found: {items}".
                                  format(items=results))
                    sys.exit(1)
                # make sure the templates are lists
                if not isinstance(stories["templates"], list):
                    logging.error("The templates in stories section must be lists not {type}. Exiting...".
                                  format(type=type(stories["templates"])))
                    sys.exit(1)
                # validate the templates
                self._validate_shap_templates(stories, pattern)
            return shap_narrative_stories
        except KeyError:
            return {'target': {'generic': None, 'templates': []}}

    def _validate_shap_templates(self, stories, pattern):
        for item in stories["templates"]:
            # make sure "features" and "narrative" exist in all dictionaries in the templates list
            validate_subset_list(parent_list=["features", "narrative"], parent_name="required items in templates",
                                 subset_list=item.keys(), subset_name="provided items")
            # make sure the exclude_cols and features columns don't have an intersection
            validate_intersection_cols(cols1=item["features"].keys(), cols2=self.shap_exclude_cols,
                                       cols1_name="features columns", cols2_name="exclude_cols")
            # make sure if the features are defined, the narrative is not missing
            if len(item["features"]) > 0 and "narrative" not in item:
                logging.error('The "features" are defined in an item in the templates but "narrative" is missing. '
                              'Please update the configuration file. Exiting...')
                sys.exit(1)
            # make sure the values for each column in features in positive or negative
            acceptable_values = {"p": "positive", "n": "negative"}
            for feature, value in item["features"].items():
                if isinstance(value, str):
                    feature_value = value.lower()
                    if feature_value in acceptable_values.values():
                        item["features"][feature] = {"impact": feature_value, "include_mean": True, "values": []}
                    else:
                        try:
                            item["features"][feature] = {"impact": acceptable_values[feature_value[0]], "include_mean": True, "values": []}
                        except KeyError:
                            logging.error('The value for feature {feature} should be ("positive" or "negative"). '
                                          'The value "{value}" is not understood'.format(feature=feature, value=value))
                            sys.exit(1)

                elif isinstance(value, dict):
                    validate_subset_list(parent_list=["impact", "include_mean", "values"],
                                         parent_name="required items in templates features",
                                         subset_list=value.keys(), subset_name="provided items")

                    # get the values for values, include_mean, and shap or use the defaults
                    item["features"][feature]["values"] = value.get("values", [])
                    item["features"][feature]["include_mean"] = value.get("include_mean", True)
                    item["features"][feature]["impact"] = value.get("impact", None)
                    # make sure the provided values have the correct type
                    assert isinstance(value["values"], list), "the values field for feature {feature} must be a list".format(feature=feature)
                    assert isinstance(value["include_mean"], bool), "the include_mean for feature {feature} must be boolean".format(feature=feature)
                    assert isinstance(value["impact"], str), "the shap field for feature {feature} must be string".format(feature=feature)
                    # try to infer the value for the shap field if the provided value is not acceptable
                    if value["impact"].lower() not in acceptable_values.values():
                        try:
                            item["features"][feature]["impact"] = acceptable_values[value["impact"][0]]
                        except KeyError:
                            logging.error('The impact field for feature {feature} should be ("positive" or "negative"). '
                                          '"{value}" is not understood'.format(feature=feature, value=value["impact"]))
                            sys.exit(1)

                else:
                    logging.error("The value for features items {feature} should be a string or a dictionary, "
                                  "found {type}".format(feature=feature, type=type(value)))
                    sys.exit(1)

    def get_shape_narrative_method(self):
        try:
            method = self._json_config["analysis"]["shap_params"]["narrative_params"]["method"]
            validate_subset_list(parent_list=SUPPORTED_NARRATIVE_METHODS, parent_name="valid narrative methods",
                                 subset_list=[method], subset_name="provided narrative method")
            return method
        except KeyError:
            return "order"

    def get_predictions_vs_actuals_activate(self):
        try:
            activate = self._json_config["analysis"]["predictions_vs_actuals_params"]["activate"]
            if activate:
                validate_subset_list(parent_list=SUPPORTED_ENGINE_FOR_PREDICTIONS_VS_ACTUALS_PLOT,
                                     parent_name="acceptable engines for plotting predictions vs actuals",
                                     subset_list=[self._engine],
                                     subset_name="defined engine")
                if not self._data.validation_score_activate:
                    logging.critical("scoring_params in validation_params was not activated. Learner will not make "
                                     "predictions and therefore cannot plot predictions vs actuals. Please update "
                                     "your configuration file and try again. Exiting...")
                    sys.exit(1)
            return activate
        except KeyError:
            return False



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import sys
import re
import warnings
import logging

from learner.configuration.configuration import Configuration


class OutlierManager:
    def __init__(self, conf: Configuration, data):
        self._conf = conf
        self._data = data

    @property
    def conf(self):
        return self._conf

    @property
    def data(self):
        return self._data

    def sd_filtering(self):
        """Find the rows that fall on or outside the defined standard deviation cut-off value from the defined center
        for the selected columns. Then, depending on the method, remove or clip the values.

        :return data: filtered/updated dataframe
        """
        initial_num_rows = self._data.shape[0]

        logging.info("Performing sd_filtering...")

        drop_idx = set()
        for col, sd_params in self._conf.outlier.sd_dict.items():
            try:
                std = self._data[col].std()
                if sd_params['center'] == 'mode':
                    mid = getattr(self._data[col], sd_params['center'])()[0]
                else:
                    mid = getattr(self._data[col], sd_params['center'])()

                lower = mid - abs(sd_params['value']) * std
                upper = mid + abs(sd_params['value']) * std

                if sd_params["method"] == "remove":
                    output = self._data[(self._data[col] >= upper) | (self._data[col] <= lower)].index
                    # I don't drop data here because if I do, the standard deviation of a column can change and the
                    # calculations will be incorrect
                    drop_idx = drop_idx.union(set(output))

                # unlike "remove" method, we can immediately invoke clip because it won't affect std of other columns
                # (because we don't remove any rows)
                if sd_params["method"] == "clip":
                    self._data[col].clip(lower, upper, inplace=True)
            except TypeError:
                warnings.warn("Filtering on column {0} was unsuccessful. This could be related to "
                              "dtype".format(col), Warning)

        # drop the indices from "remove" method
        self._data = self._data.loc[~self._data.index.isin(drop_idx)]
        self.print_filtering_info(initial_num_rows, "sd_filtering")
        return self._data

    def min_max_filtering(self):
        """Find the rows that fall outside the "min" and "max" limits. Then, depending on the method, remove the entire
        row or clip the values in the columns.

        :return data: filtered/updated dataframe
        """
        initial_num_rows = self._data.shape[0]

        logging.info("Performing min_max operations...")

        for col, min_max_params in self.conf.outlier.min_max_dict.items():
            try:
                if min_max_params["method"] == "remove":
                    index = (self._data[col] < min_max_params["min"]) | (self._data[col] > min_max_params["max"])
                    self._data = self._data.loc[~index]

                if min_max_params["method"] == "clip":
                    self._data[col].clip(min_max_params["min"], min_max_params["max"], inplace=True)
            except TypeError:
                warnings.warn("Filtering on column {0} was unsuccessful. This could be related to "
                              "dtype".format(col), Warning)

        self.print_filtering_info(initial_num_rows, "min_max_filtering")

        # this line is unnecessary but helps with testing
        return self._data

    def quantile_filtering(self):
        """Find the rows that fall outside a minimum quantile and a maximum quantile. Then, depending on the method,
        remove the entire row or clip the values in the columns.

        :return data: filtered/updated dataframe
        """
        initial_num_rows = self._data.shape[0]
        logging.info("Performing quantile operations...")
        for col, quantile_params in self.conf.outlier.quantile_dict.items():
            try:
                lower = self._data[col].quantile(quantile_params["min"]) if quantile_params["min"] is not None else None
                upper = self._data[col].quantile(quantile_params["max"]) if quantile_params["max"] is not None else None
                if quantile_params["method"] == "remove":
                    index = (self._data[col] < lower) | (self._data[col] > upper)
                    self._data = self._data.loc[~index]

                if quantile_params["method"] == "clip":
                    self._data[col].clip(lower, upper, inplace=True)
            except TypeError:
                warnings.warn("Filtering on column {0} was unsuccessful. This could be related to "
                              "dtype".format(col), Warning)

        self.print_filtering_info(initial_num_rows, "quantile_filtering")

        # this line is unnecessary but helps with testing
        return self._data

    def value_filtering(self):
        """Loop over the columns that require value filtering or replacements. Remove the rows or replace the value based
        on the items in value_dict.

        :return data: filtered/updated dataframe
        """
        initial_num_rows = self._data.shape[0]
        logging.info("Performing value operations...")
        for col, col_info in self.conf.outlier.value_dict.items():
            for to_replace, replace_with in col_info.items():
                try:
                    pattern = re.compile(to_replace)
                    if replace_with is None:
                        index = self._data[col].str.match(pattern)
                        self._data = self._data[~index]
                    else:
                        pass
                        self._data.loc[:, col] = self._data.loc[:, col].str.replace(pattern, replace_with)
                except (AttributeError, TypeError):
                    warnings.warn(f"Value filtering Filtering on column {col} was unsuccessful. This could be related to "
                                  f"the data type of the column", Warning)

        self.print_filtering_info(initial_num_rows, "value_filtering")
        # this line is unnecessary but helps with testing
        return self._data

    def print_filtering_info(self, initial_num_rows, method_name):
        """Print useful information before and after performing outlier filtering


        :param initial_num_rows: the initial number of rows in the training data
        :param method_name: the method name to create the output message
        :return: None
        """
        # compute how many rows are left and print some useful information
        final_num_rows = self._data.shape[0]
        filtered_num_rows = initial_num_rows - final_num_rows
        logging.info("%i rows were filtered during %s and %i rows are left", filtered_num_rows, method_name,
                     final_num_rows)
        # exit if no data is left for training after filtering
        if final_num_rows == 0:
            logging.error("No data is left for training after %s. Exiting...", method_name)
            sys.exit(1)

    def handle_outlier(self):
        """The main function for filtering the data and removing outliers. Depending on the user input, appropriate
        if blocks will execute.

        :return: updated data
        """

        logging.info("Handling outliers in training data...")
        if self._conf.outlier.quantile_activate:
            self.quantile_filtering()

        if self._conf.outlier.sd_activate:
            self.sd_filtering()

        if self._conf.outlier.min_max_activate:
            self.min_max_filtering()

        if self._conf.outlier.value_activate:
            self.value_filtering()

        return self._data



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""Handle segmenting a dataset based on the values of a specific column"""
import sys
import abc
import warnings
import logging
import numpy as np
import pandas as pd

from learner.configuration.configuration import Configuration


class AbstractSegmenter:
    """A template class for other segmenter classes."""
    __metaclass__ = abc.ABCMeta

    def __init__(self, conf: Configuration):  # pragma: no cover
        self._conf = conf
        self._seg_list = conf.segmenter.seg_list
        self._max_num_bins = conf.segmenter.max_num_bins
        self.seg_dict = None

    @property
    def conf(self):
        return self._conf

    @property
    def seg_list(self):
        return self._seg_list

    @property
    def max_num_bins(self):
        return self._max_num_bins

    @staticmethod
    def fit(data, train_col):  # pragma: no cover
        pass

    @staticmethod
    def transform(data, col, name):  # pragma: no cover
        pass

    @staticmethod
    def fit_transform(data, train_col, train_name):  # pragma: no cover
        pass

    def create_seg_dict(self):
        """Create a dictionary of the segments from seg_list where the key is the segment border and the value is the
        segment id. This dictionary is used to update the model_dict and enable custom parameters/model for each
        segment.

        :return: None
        """
        self.seg_dict = {seg_border: seg_id for seg_id, seg_border in enumerate(self.seg_list)}

    def update_models_dict(self):
        """Make some updates to the models_dict by finding the seg_id and seg_name for each segment. Then reformat the
        models_dict in a way that the value of each tag will be a dictionary of dictionaries. The key of each item of
        the parent dictionary is the seg_id, and the value is a dictionary that contain all the parameters. Here's an
        example:
            {'rf1': {0: {'params': {'min_samples_leaf': 1,
                                    'min_samples_split': 4,
                                    'n_estimators': 100,
                                    'n_jobs': -1},
                         'seg_id': 0,
                         'type': 'RandomForestClassifier'},
                     1: {'params': {'min_samples_leaf': 1,
                                    'min_samples_split': 4,
                                    'n_estimators': 100,
                                    'n_jobs': -1},
                         'path': '/home/ramezani/dev/output/classifier/classifier_rf1_2019-06-20-22-14/',
                         'seg_id': 1,
                         'type': 'XGBClassifier'}}
        :return: None
        """
        self.create_seg_dict()
        self._convert_dict_to_list()
        self._set_seg_ids()
        self._reformat_models_dict()
        self._ensure_all_segs_have_params()

    def _convert_dict_to_list(self):
        """The user has the option to pass a dictionary or a list of dictionaries in the models_dict field. If a
        dictionary is provided, it indicates that all segments will use similar parameters (no custom parameter).
        In that case, we need to set seg_id to -1 and seg_name to "any" and then convert the dictionary to a list
        for consistency.

        :return:  None
        """
        # loop through all items in models_dict, find the ones that are dictionaries and make necessary updates.
        for tag, mdls in self.conf.model.models_dict.items():
            if isinstance(mdls, dict):
                self.conf.model.models_dict[tag]["seg_name"] = "any"
                self.conf.model.models_dict[tag]["seg_id"] = -1
                self.conf.model.models_dict[tag] = [mdls]

    def _set_seg_ids(self):
        """Find and set seg_id for all items in the models_dict. Also, make sure seg_name and seg_id are consistent in
        case both of them are provided.

        :return: None
        """
        for tag, mdls in self.conf.model.models_dict.items():
            for i, mdl in enumerate(mdls):
                # if seg_id is not defined, try to figure things out using seg_name
                if "seg_id" not in mdl:
                    try:
                        seg_name = mdl["seg_name"]
                        mdl["seg_id"] = self.seg_dict[seg_name] if seg_name != "any" else -1
                    except KeyError:
                        mdl["seg_id"] = -1
                        mdl["seg_name"] = "any"
                elif mdl["seg_id"] == -1:
                    mdl["seg_name"] = "any"
                # if both seg_id and seg_name are defined, make sure they both are correct
                elif "seg_id" in mdl and "seg_name" in mdl:
                    try:
                        if mdl["seg_id"] != self.seg_dict[mdl["seg_name"]]:
                            logging.error("The seg_id '{seg_id}' and seg_name '{seg_name}' in {tag} are not consistent."
                                          " Please update the configuration file and try again. Exiting...".
                                          format(seg_id=mdl["seg_id"],
                                                 seg_name=mdl["seg_name"],
                                                 tag=tag))
                            sys.exit(1)
                    except KeyError:
                        logging.error("The seg_name '{seg_name}' is defined under {tag} but does not contain "
                                      "in the data. There could be a typo there. Please update the configuration file "
                                      "and try again. Exiting..."
                                      .format(seg_id=mdl["seg_id"],
                                              seg_name=mdl["seg_name"],
                                              tag=tag))

                        sys.exit(1)

    def _reformat_models_dict(self):
        """Reformat the models_dict and create a dictionary of dictionaries. Please see the documentation of
        update_models_dict method for more details.

        :return: None
        """
        for tag, mdls in self.conf.model.models_dict.items():
            # check if default parameters exist
            default_exist = any(is_default for is_default in [(mdl.get("seg_id", None) == -1 or mdl.get("seg_name", None) == "any") for mdl in mdls])
            # if there are default parameters, inform the user that learner will use them
            if len(mdls) < len(self.seg_list) and default_exist:
                logging.info("The default parameters have been provided. Learner will use them for segments without "
                             "customized parameters")
            if len(mdls) < len(self.seg_list) and not default_exist:
                logging.error("The number of segments are {num_segments} but {num_params} parameters was provided "
                              "without providing a default parameter. Please update the configuration file. Exiting".
                              format(num_segments=len(self.seg_list), num_params=len(mdls)))
                sys.exit(1)
            self.conf.model.models_dict[tag] = {mdl["seg_id"]: mdl for mdl in mdls}

    def _ensure_all_segs_have_params(self):
        """We need to make sure the number of items in models_dict matches the number of segments. If the number of
        parameters do not match, we need to use default values for the segments that we don't have a custom parameter
        for. This method takes care of that. It will raise an error if we need a default parameter but it hasn't been
        provided or there are other inconsistencies. This is the last step in updating the models_dict.

        :return: None
        """
        for seg_id, seg_border in enumerate(self.seg_list):
            for tag, mdls in self.conf.model.models_dict.items():
                if seg_id not in self.conf.model.models_dict[tag]:
                    try:
                        self.conf.model.models_dict[tag][seg_id] = self.conf.model.models_dict[tag][-1]
                    except KeyError:
                        logging.error("The parameters set for different segments are incorrect. There are probably "
                                      "overlaps between seg_name and seg_id with no default parameters")
                        sys.exit(1)


class SegmenterHandler:
    """SegmenterHandler is the workflow class that is responsible for instantiation and running segmentation."""

    def __init__(self, conf, data_type="train"):
        """Initialize a SegmenterHandler object using a conf object and initialize the instance variables.

        :param conf: a conf object
        """
        self._conf = conf
        self._data_type = data_type

    @property
    def conf(self):
        return self._conf

    def handle_segmenter(self, data):
        """Execution method for controlling the creation and transformation of data with a segmenter

        :param data: a pandas dataframe
        :return: None
        """
        # if segmenter is not activated immediately return, otherwise continue
        if not self._conf.segmenter.activate:
            return
        segmenter = None
        seg_type = self._conf.segmenter.type
        # find the segmenter type and instantiate the corresponding object
        if seg_type == 'static':
            segmenter = StaticSegmenter(self._conf)
        elif seg_type == 'dynamic' and self.conf.segmenter.duplicates == "merge":
            segmenter = DynamicSegmenter(self._conf)
        elif seg_type == 'dynamic' and self.conf.segmenter.duplicates == "drop":
            segmenter = QCutSegmenter(self._conf)
        elif seg_type == 'value':
            segmenter = ValueSegmenter(self._conf)
        # if seg_list hasn't been populated (commonly during training phase) go through the entire process of fitting,
        # transforming, and populating it.
        if not self._conf.segmenter.seg_list:

            segmenter.fit_transform(data,
                                    self._conf.segmenter.train_col,
                                    self._conf.segmenter.train_name)

            segmenter.transform(data,
                                self._conf.segmenter.test_col,
                                self._conf.segmenter.test_name)

            self._conf.segmenter.seg_list = segmenter.seg_list
        # if the seg_list has already been populated, then just do the transform
        else:
            # we need to do this because the test_col might be precessed, i.e. label encoding, etc
            if seg_type == "value" and self._data_type == "validation":
                segmenter.transform(data,
                                    self._conf.segmenter.train_col,
                                    self._conf.segmenter.test_name)
            else:
                segmenter.transform(data,
                                    self._conf.segmenter.test_col,
                                    self._conf.segmenter.test_name)


class ValueSegmenter(AbstractSegmenter):
    """Segment a column by finding the unique value in that column. This is suitable for categorical columns or columns
     with limited number of unique values. This is a pretty straightforward segmenter. We just get a list of unique
     values to create the seg_list, and then copy the columns. The processor and configuration classes perform some
     operations to enable an easy implementation here. For example, if a column is a categorical column and gets
     label-encoded, the pre-encoded column needs to be used while processing the training data because the test data
     get's loaded before processing. The processor makes a backup and configuration updates the column names"""

    def __init__(self, conf):
        """Initialize a segmenter object using a conf object.

        :param conf: an instance of the Configuration class
        """
        super().__init__(conf)

    def fit(self, data, train_col):
        """Get a list that contain all the unique values in the train_col. train_col is an existing column in the data.

        :param data: pandas DataFrame
        :param train_col: the column to use for segmentation
        :return: a list containing the upper boundary of the segments
        """
        # we first try to sort the list, if we give up and continue with unsorted list if we fail. This can happen when
        # there are missing values in segment column and the get imputed with a number.
        try:
            self._seg_list = sorted(data[train_col].unique().tolist())
        except Exception:
            self._seg_list = data[train_col].unique().tolist()
        self.update_models_dict()

    def transform(self, data, col, name):
        """Transform a column with a trained segment list. This function copies the column names "col" into a new
        column named "name"

        :param data: pandas DataFrame
        :param col: column to use
        :param name: name of the new column
        """
        data[name] = data[col]

    def fit_transform(self, data, train_col, train_name):
        """Create a segment list and transform the same column.

         :param data: pandas DataFrame
         :param train_col: the columns to be used for segmentation
         :param train_name: name of column to assign transformed column
         :return: None
         """
        self.fit(data, train_col)
        self.transform(data, train_col, train_name)


class StaticSegmenter(AbstractSegmenter):
    """Segment a column by the bin size and max number of bins specified."""

    def __init__(self, conf):
        """Initialize a segmenter object using a conf object.

        :param conf: an instance of the Configuration class
        """
        super().__init__(conf)
        self._bin_size = conf.segmenter.bin_size

    @property
    def bin_size(self):
        return self._bin_size

    def _get_segment_id(self, value):
        """Given a value, return the segment id for that value.

        :param value: a numeric value
        :return: an integer corresponding to segment id
        """
        return int(min(value // self._bin_size, self._max_num_bins - 1))

    def _get_seg_border(self, value):
        """Compute the upper boundary of a segment using bin_size and segment id

        :param value: a number to find the segment border for
        :return: upper boundary of a segment for the value
        """
        seg_id = self._get_segment_id(value)
        return (seg_id + 1) * self._bin_size

    def fit(self, data, train_col):
        """Create a column with label name that contains the segment border for each entry. Use an existing column in
        the data set (train_col) for calculations.

        :param data: pandas DataFrame
        :param train_col: the column to use for segmentation
        :return: a list containing the upper boundary of the segments
        """
        segment = data[train_col].apply(lambda x: self._get_seg_border(x))
        self._seg_list = sorted(segment.unique().tolist())
        self.update_models_dict()

    def transform(self, data, col, name):
        """Transform a column with a trained segment list. This function creates a new column in the dataset (called name),
         which contain the upper boundary of the segment corresponding the the values in col.

        :param data: pandas DataFrame
        :param col: column to use for finding the segments
        :param name: name of column to assign transformed column
        """
        data[name] = data[col].apply(lambda x: self._get_seg_border(x))

    def fit_transform(self, data, train_col, train_name):
        """Create a segment list and transform the same column.

         :param data: pandas DataFrame
         :param train_col: the columns to be used for segmentation
         :param train_name: name of column to assign transformed column
         :return: None
         """
        self.fit(data, train_col)
        self.transform(data, train_col, train_name)


class DynamicSegmenter(AbstractSegmenter):
    """Creates segments that are dynamic in width based on the number of items in each segment."""

    def __init__(self, conf):
        """Initialize a DynamicSegmenter instance using a conf object.

        :param conf: a instance of the Configuration class
        """
        super().__init__(conf)

    def _get_seg_borders(self, data, train_col):
        """Obtain the segment borders (both upper and lower boundaries) for train_col in the data

        :param data: a pandas DataFrame
        :param train_col: the column in the dataset to use for finding the segment borders
        :return: a list of lists containing the lower and upper boundaries of the segments
        """
        seg_borders = []
        delta = int(data.shape[0]/self._max_num_bins)
        lower = 0
        upper = delta - 1
        sorted_train_col = np.sort(data[train_col])
        for i in range(self._max_num_bins):
            seg_borders.append([sorted_train_col[lower], sorted_train_col[upper]])
            lower = upper + 1
            upper = lower + delta - 1
        # the last border should be the maximum value in the column
        seg_borders[-1][1] = sorted_train_col[-1]
        return seg_borders

    @staticmethod
    def _merge_seg_borders(seg_borders):
        """If upper boundary of a segment is the same as the lower boundary of the next segment, merge those intervals.

        :param seg_borders: a list of lists containing the lower and upper boundaries of the segments
        :return: a list of lists containing the merged segment borders
        """
        merged_seg_borders = []
        for border in seg_borders:
            if not merged_seg_borders or merged_seg_borders[-1][1] < border[0]:
                merged_seg_borders.append(border)
            else:
                merged_seg_borders[-1][1] = max(merged_seg_borders[-1][1], border[1])
        return merged_seg_borders

    def fit(self, data, train_col):
        """Create and store the upper boundaries for segments

        :param data: pandas DataFrame
        :param train_col: the column to be used for segmentation fitting
        """
        logging.info("Attempting to split the training dataset into %i segment(s)", self._max_num_bins)
        seg_borders = self._get_seg_borders(data, train_col)
        merged_seg_borders = self._merge_seg_borders(seg_borders)
        self._seg_list = [border[1] for border in merged_seg_borders]

        if len(self._seg_list) < self._max_num_bins:
            warnings.warn("The training dataset was split into {0} segments due to overlaps in the boundaries"
                          .format(len(self._seg_list)), Warning)
        else:
            logging.info("The training dataset was successfully split into %i segment(s)", self._max_num_bins)
        self.update_models_dict()

    def transform(self, data, col, name):
        """Transform a column with a trained segment list.

        :param data: pandas DataFrame
        :param col: column to be used for segmentation transformation
        :param name: column name for the output of segmentation transformation
        :return: None
        """
        seg_list = self._seg_list

        # This is done so that the edges take into account lower and higher values in a prediction dataset than were
        # seen in the training dataset.
        seg_list = [-np.inf] + seg_list

        if data[col].max() > max(self._seg_list):
            max_value = data[col].max()
            seg_list[-1] = max_value

        data[name] = pd.cut(data[col], seg_list, labels=self._seg_list, right=True)

    def fit_transform(self, data, train_col, train_name):
        """Create a segment list and transform the same column.

        :param data: pandas DataFrame
        :param train_col: the column to be used for segmentation fitting
        :param train_name: the column name for the output of segmentation transformation
        :return: None
        """
        self.fit(data, train_col)
        self.transform(data, train_col, train_name)


class QCutSegmenter(AbstractSegmenter):
    """Creates segments that are dynamic in width based on the number of items in each segment. The difference between
    the DynamicSegment and the QCutSegmenter is how the boundaries are defined. Here we use pandas's qcut."""

    def __init__(self, conf):
        """Initialize a QCutSegmenter instance using a conf object.

        :param conf: a instance of the Configuration class
        """
        super().__init__(conf)

    def _get_seg_borders(self, data, train_col):
        """Obtain the segment borders for train_col in the data. Here, we get a single list that contains the borders
        instead of list of lists. This is provided by pandas' qcut method. For more information, see
        https://pandas.pydata.org/docs/reference/api/pandas.qcut.html

        :param data: a pandas DataFrame
        :param train_col: the column in the dataset to use for finding the segment borders
        :return: a list containing the lower and upper boundaries of the segments
        """
        _, seg_borders = pd.qcut(data[train_col], self._max_num_bins, duplicates="drop", retbins=True)

        # we ignore the first one to later replace it with -inf
        return list(seg_borders[1:]) if len(seg_borders) != 1 else list(seg_borders)

    def fit(self, data, train_col):
        """Create and store the upper boundaries for segments

        :param data: pandas DataFrame
        :param train_col: the column to be used for segmentation fitting
        """
        logging.info("Attempting to split the training dataset into %i segment(s)", self._max_num_bins)
        self._seg_list = self._get_seg_borders(data, train_col)

        if len(self._seg_list) < self._max_num_bins:
            warnings.warn("The training dataset was split into {0} segments due to overlaps in the boundaries"
                          .format(len(self._seg_list)), Warning)
        else:
            logging.info("The training dataset was successfully split into %i segment(s)", self._max_num_bins)
        self.update_models_dict()

    def transform(self, data, col, name):
        """Transform a column with a trained segment list.

        :param data: pandas DataFrame
        :param col: column to be used for segmentation transformation
        :param name: column name for the output of segmentation transformation
        :return: None
        """
        seg_list = self._seg_list.copy()

        # This is done so that the edges take into account lower and higher values in a prediction dataset than were
        # seen in the training dataset.
        seg_list = [-np.inf] + seg_list

        if data[col].max() > max(self._seg_list):
            max_value = data[col].max()
            seg_list[-1] = max_value

        data[name] = pd.cut(data[col], seg_list, labels=self._seg_list, right=True)

    def fit_transform(self, data, train_col, train_name):
        """Create a segment list and transform the same column.

        :param data: pandas DataFrame
        :param train_col: the column to be used for segmentation fitting
        :param train_name: the column name for the output of segmentation transformation
        :return: None
        """
        self.fit(data, train_col)
        self.transform(data, train_col, train_name)



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module for sampling the data. Currently, the main functionality include handling the imbalanced data for
classification problems using over or under-sampling techniques. This module should also handle other sampling related
tasks. Please note that, sampling the data for loading the training dataset is handled by data_loader module."""
import sys
import logging
import pandas as pd
from sklearn.model_selection import train_test_split

from learner.configuration.configuration import Configuration
from learner.data_worker.output_handler import OutputHandler


class DataSampler(object):
    """The main class to handle data_sampling tasks. The main driver method here is the sample_data."""
    def __init__(self, conf: Configuration, data: pd):
        """Instantiate a DataSampler object by providing a conf object and a reference to the training dataset.

        :param conf: an instance of the Configuration class
        :param data: a reference to the training dataset
        """
        self._conf = conf
        self.data = data
        # this hold the number of items in the majority class
        self.majority_count = None
        # this will be the label for the majority class (true or false, 0 or 1)
        self.majority_label = None
        # this will be the indices (true or false list) to identify the rows for the majority class
        self.majority_rows = None

        # the same as above but for the minority class
        self.minority_count = None
        self.minority_label = None
        self.minority_rows = None

        # these are the train and test indices if train_test_split was activated
        self.train_indices = None
        self.test_indices = None

    @property
    def conf(self):
        return self._conf

    def get_classes(self):
        """Get the information about the classes in the target columns. This is done by calling the value_counts method
        on the target column.

        :return: None
        """
        classes_df = self.data[self.conf.column.target_col].value_counts()
        num_classes = classes_df.shape[0]
        if num_classes != 2:
            logging.error("Learner only supports over/undersampling for binary classifications. Exiting...")
            sys.exit(1)
        # the class_df will be sorted, that's why this logic works
        self.majority_count = classes_df.iloc[0]
        self.majority_label = classes_df.index[0]
        self.minority_count = classes_df.iloc[1]
        self.minority_label = classes_df.index[1]

        logging.info(f"There are {self.majority_count} items for class {self.majority_label} and {self.minority_count} "
                     f"for class {self.minority_label}")

        # get the rows for the minority class
        self.minority_rows = self.data[self.conf.column.target_col] == self.minority_label
        self.majority_rows = self.data[self.conf.column.target_col] == self.majority_label

    def oversampling_handler(self):
        """Oversample the minority class to have as many as majority instances in the training data. To make this happen
        the minority class is being sampled with replacement. Note that this method does not add any noise to the data.
        That means the minority row are being duplicated in the training data.

        :return: None
        """
        # upsample the minority class to have as many as majority class instances (note that this only duplicates the
        # data)
        logging.info("Upsampling the minority class...")
        self.data = pd.concat([self.data[self.majority_rows],
                               self.data[self.minority_rows].sample(self.majority_count,
                                                                    replace=True,
                                                                    random_state=self.conf.sample.imbalanced_random_state)],
                              ignore_index=True)

    def undersampling_handler(self):
        """Under-sample the majority class to have as few as the minority class in the training dataset. Here, the
        majority rows are being sampled without replacement.

        :return: None
        """
        # sample the majority class (as many as the minority) and concatenate with minority
        logging.info("Undersampling the majority class...")
        self.data = pd.concat([self.data[self.majority_rows].sample(self.minority_count,
                                                                    replace=False,
                                                                    random_state=self.conf.sample.imbalanced_random_state),
                               self.data[self.minority_rows]],
                              ignore_index=True)

    def handle_random_train_test_split(self):
        """Randomly split the original train data into the new train and test data using the parameters provided. Then
        write the test data to disk so that Learner will use it when making predictions. Also, if requested, write the
        new train data to disk so that user can access it if needed.

        :return: None
        """
        logging.info("The number of rows for the original train_data: {nrows}".format(nrows=self.data.shape[0]))
        try:
            data, test_data = train_test_split(self.data,
                                               test_size=self.conf.sample.split_test_size,
                                               random_state=self.conf.sample.split_random_state,
                                               shuffle=self.conf.sample.split_shuffle,
                                               stratify=self.data[self.conf.sample.split_random_stratify]
                                               if self.conf.sample.split_random_stratify else None)

            self.test_indices = test_data.index
            self.train_indices = data.index
            if self.conf.sample.split_data:
                self.data = data
            self._write_split_data(test_data)
        except Exception as e:
            logging.error("Splitting the training data was unsuccessful. The error is {error}. Exiting".format(error=e))
            sys.exit(1)

    def handle_sort_train_test_split(self):
        """Split the training data into two date sets to use for training and testing using a date column.
        To make this work, we first convert the date column to datetime type. We then call the appropriate methods
        depending on the user input. If test_on_after is defined, we select the train and test data based on the dates
        provided by user otherwise we sort the training data by the date column and then split it.

        :return: None
        """
        nrows = self.data.shape[0]
        logging.info(f"The number of rows for the original train_data: {nrows}")
        try:
            # we need to convert the sort_col to a date type
            self.data[self.conf.sample.split_sort_col] = pd.to_datetime(self.data[self.conf.sample.split_sort_col],
                                                                        errors='coerce',
                                                                        infer_datetime_format=True)
            if self.conf.sample.split_test_on_after:
                data, test_data = self._handle_sort_split_using_dates()
            else:
                data, test_data = self._handle_sort_split_using_test_size()

            # if the user wants to shuffle train data, we do it here
            if self.conf.sample.split_shuffle:
                data = data.sample(frac=1, random_state=self.conf.sample.split_random_state)

            self.test_indices = test_data.index
            self.train_indices = data.index
            if self.conf.sample.split_data:
                self.data = data

            self._write_split_data(test_data)
        except Exception as e:
            logging.error("Splitting the training data was unsuccessful. The error is {error}. Exiting".format(error=e))
            sys.exit(1)

    def _handle_sort_split_using_dates(self):
        """Assuming that two dates are defined, we split the training data into two dataset. If one of those datasets
        are empty after the split, we issues an error.

        :return: the new training dataset and the test dataset
        """
        test_data = self.data[self.data[self.conf.sample.split_sort_col] >= self.conf.sample.split_test_on_after]
        data = self.data[self.data[self.conf.sample.split_sort_col] < self.conf.sample.split_train_before]
        if data.empty:
            logging.critical("""No data is left for training after train_test_split. Please update the configuration  
            file and try again. Exiting...""")
            sys.exit(1)

        if test_data.empty:
            logging.critical("""No data is left for testing after train_test_split. Please update the configuration  
            file and try again. Exiting...""")
            sys.exit(1)

        return data, test_data

    def _handle_sort_split_using_test_size(self):
        """Sort the training data set and split it into two datasets for training and testing. Issue an error if
        sampling leads to an empty dataset.

        :return: the new training dataset and the test dataset
        """
        nrows = self.data.shape[0]
        self.data.sort_values(by=self.conf.sample.split_sort_col,
                              ascending=True,
                              inplace=True,
                              na_position=self.conf.sample.split_sort_nan_position)

        # compute the number of rows we need for splitting
        # when split_test_size < 1, we need to get a fraction of data
        if self.conf.sample.split_test_size < 1:
            num_test_rows = int(self.data.shape[0] * self.conf.sample.split_test_size)
        else:
            num_test_rows = self.conf.sample.split_test_size

        data, test_data = (self.data.iloc[:nrows - num_test_rows, :],
                           self.data.iloc[nrows - num_test_rows:, :])
        if num_test_rows >= nrows:
            logging.critical("""test_size in train_test_split is greater than the number of rows in train data.
            Please update you configuration file by choosing a smaller number and try again. Exiting...""")
            sys.exit(1)
        return data, test_data

    def _write_split_data(self, test_data):
        """Write the test data to disk so that Learner can use it when making predictions. Also, if requested, write the
        new train data to disk so that user can access it if needed.

        :param test_data: the test_data to save to disk
        :return: None
        """
        logging.info("The number of rows for the train data after split: {nrows}".format(nrows=self.data.shape[0]))
        logging.info("The number of rows for the test data after split: {nrows}".format(nrows=test_data.shape[0]))

        logging.info("Writing the split test data in {filepath}"
                     .format(filepath=self.conf.workspace.data_directory + self.conf.data.sample_validation_filename))
        out = OutputHandler(self.conf)
        out.save_file(self.conf.workspace.data_directory,
                      self.conf.data.sample_validation_filename,
                      test_data,
                      add_timetag=False,
                      format=self.conf.data.validation_format,
                      sep=self.conf.data.test_delimiter)

        if self.conf.sample.split_save_train_data:
            logging.info("Writing the split train data in {filepath}"
                         .format(filepath=self.conf.workspace.data_directory + self.conf.data.sample_train_filename))

            out.save_file(self.conf.workspace.data_directory,
                          self.conf.data.sample_train_filename,
                          self.data,
                          add_timetag=False,
                          format=self.conf.data.train_format,
                          sep=self.conf.data.train_delimiter)

    def sample_data(self):
        """This is the main driver method for the DataSampler class. In general, this method is called after
        instantiating the object.

        :return: a reference to the training data
        """
        if self.conf.sample.split_activate and self.conf.sample.split_method == "random":
            self.handle_random_train_test_split()

        if self.conf.sample.split_activate and self.conf.sample.split_method == "sort":
            self.handle_sort_train_test_split()

        if self.conf.sample.imbalanced_activate:
            logging.info("Balancing the training data...")
            # we first populate the information on each class
            self.get_classes()
            getattr(self, self.conf.sample.imbalanced_method + "_handler")()
        return self.data



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""Handles all operations related to saving the output and the models including pickled models"""

import sys
import os
import pickle
import warnings
import logging
import pandas as pd
import numpy as np

from learner.utilities.templates import (PRED_PATH, PRED_PATH_SEG,
                                         PRED_PATH_CHUNK, PRED_PATH_CHUNK_SEG)


class OutputHandler:
    def __init__(self, conf, data_type=""):
        self._conf = conf
        self.data_type = data_type
        if data_type:
            self.dtype_sep = f"_{data_type}_"
        else:
            self.dtype_sep = "_"

    @property
    def conf(self):
        return self._conf

    def save_file(self, directory_path, filename, dataframe, add_timetag=True, index=False, format="csv",
                  add_version=False, version_name="", version_column_name="version",
                  **kwargs):
        """Save a dataframe into the hard drive and create a timetag column if necessary. This method accepts a "format"
         argument and calls the appropriate methods to support multiple file formats.
        formats including

        :param directory_path: the path to the folder in which the file should be saved.
        :param filename: the name of the file.
        :param dataframe: the dataframe that contains the data.
        :param add_timetag: a boolean to define if a timetag should be added or not given add_timetag is true in conf object.
        :param index: a boolean to specify if the row names should be written or not
        :param format: the file format when saving the data, currently we support "csv", "parquet", and "feather"
        :param add_version: whether we should add the version column or not
        :param version_name: the value to use in the version column
        :param version_column_name: the name of the version column
        :return: None
        """
        assert isinstance(dataframe, pd.DataFrame), "ERROR: save_file only supports saving dataframes"
        if add_timetag:
            timetag = pd.Series(np.full(dataframe.shape[0], self._conf.timetag), name="timetag")
            dataframe = pd.concat((dataframe, timetag), axis=1)
        if add_version:
            dataframe[version_column_name] = version_name

        if format == "csv":
            self.save_csv(directory_path, filename, dataframe, index=index, **kwargs)
        elif format == "parquet":
            self.save_parquet(directory_path, filename, dataframe, index=index)
        elif format == "feather":
            self.save_feather(directory_path, filename, dataframe)

    def save_csv(self, directory_path, filename, dataframe, index=False, **kwargs):
        """Save a dataframe into the hard drive in the csv format.

        :param directory_path: the path to the folder in which the file should be saved.
        :param filename: the name of the file.
        :param dataframe: the dataframe that contains the data.
        :param index: a boolean to specify if the row names should be written or not
        :return: None
        """
        dataframe.to_csv(directory_path + filename, index=index, **kwargs)

    def save_parquet(self, directory_path, filename, dataframe, index=False):
        """Save a dataframe into the hard drive in the parquet format.

        :param directory_path: the path to the folder in which the file should be saved.
        :param filename: the name of the file.
        :param dataframe: the dataframe that contains the data.
        :param index: a boolean to specify if the row names should be written or not
        :return: None
        """
        try:
            dataframe.to_parquet(directory_path + filename, engine="pyarrow", index=index)
        except Exception as e:
            logging.error(f"Failed to save the data. The error is {str(e)}. Will try a different engine.")
            dataframe.to_parquet(directory_path + filename, engine="fastparquet", index=index)

    def save_feather(self, directory_path, filename, dataframe):
        """Save a dataframe into the hard drive into the feather format.

        :param directory_path: the path to the folder in which the file should be saved.
        :param filename: the name of the file.
        :param dataframe: the dataframe that contains the data.
        :return: None
        """
        # to resolve the error: feather does not support serializing a non-default index for the index;
        # you can .reset_index() to make the index into column(s)
        dataframe.reset_index(inplace=True, drop=True)
        dataframe.to_feather(directory_path + filename)

    def concat_save_csv(self, models, items, output_name, sep_timetag):
        """Read a series of csv file from the disk, concatenate them, and write the output to a csv file

        :param models: a dictionary of models that contains tag information
        :param items: a list containing the items (this is seg_list for the recommender engine)
        :param output_name: the prefix of the filenames (this is usually conf.workspace.name)
        :param sep_timetag: the sep_timetag for the model
        :return: None
        """
        logging.info("Combining the results for different segments")
        # read and concatenate the results
        for tag, mdls in models.items():
            filename = PRED_PATH.format(
                path=mdls.get(0, mdls)["path"],
                output_name=self._conf.workspace.name,
                tag=str(tag),
                sep_timetag=str(sep_timetag),
                dtype_sep=self.dtype_sep
            )
            if os.path.isfile(filename):
                warnings.warn("{0} already exist. It will be overwritten...".format(filename), Warning)
                os.remove(filename)

            for seg_id in range(len(items)):
                try:
                    item_name = PRED_PATH_SEG.format(
                        path=mdls.get(seg_id, mdls)["path"],
                        output_name=output_name,
                        dtype_sep=self.dtype_sep,
                        tag=str(tag),
                        seg_id=str(seg_id),
                        sep_timetag=str(sep_timetag)
                    )
                    df = pd.read_csv(item_name)
                    logging.info("Appending %s to %s", item_name, filename)
                    if os.path.isfile(filename) is False:
                        df.to_csv(filename, index=False)
                    else:
                        df.to_csv(filename, mode='a', header=False, index=False)
                except FileNotFoundError as e:
                    logging.error(str(e))

        if self._conf.data.test_clean_up:
            self.delete_concat_save_csv(models, items, output_name, sep_timetag)

    def concat_chunk_csv(self, num_chunks, models_dict, seg_id=None):
        """Read a series of csv file from the disk, concatenate them, and write the output to a single csv file.
        This function concatenates the predictions for different chunks and different models and produces a
        single csv file for each model.

        :param num_chunks: the number of chunks to be used for reading the file indices
        :param models_dict: an item in models_dict. The value for "path" is typically used in this class
        :param seg_id: segment id
        :return: None
        """
        logging.info("Combining the predictions for different data chunks")
        for tag, mdls in models_dict.items():
            if seg_id is None:
                filename = PRED_PATH.format(
                    path=models_dict[tag]["path"],
                    output_name=self._conf.workspace.name,
                    tag=str(tag),
                    sep_timetag=str(self._conf.sep_timetag),
                    dtype_sep=self.dtype_sep
                )
            else:
                mdl = mdls.get(seg_id, mdls)
                filename = PRED_PATH_SEG.format(
                    path=mdl["path"],
                    output_name=self._conf.workspace.name,
                    tag=str(tag),
                    seg_id=str(seg_id),
                    sep_timetag=str(self._conf.sep_timetag),
                    dtype_sep=self.dtype_sep
                )

            if os.path.isfile(filename):
                warnings.warn(f"{filename} already exist. It will be overwritten...", Warning)
                os.remove(filename)

            for index in range(num_chunks):
                sep_index = f"_{index}"
                if seg_id is None:
                    chunkname = PRED_PATH_CHUNK.format(
                        path=models_dict[tag]["path"],
                        output_name=self._conf.workspace.name,
                        tag=str(tag),
                        sep_timetag=str(self._conf.sep_timetag),
                        index=str(index),
                        dtype_sep=str(self.dtype_sep)
                    )
                else:
                    mdl = mdls.get(seg_id, mdls)
                    chunkname = PRED_PATH_CHUNK_SEG.format(
                        path=mdl["path"],
                        output_name=self._conf.workspace.name,
                        tag=str(tag),
                        seg_id=str(seg_id),
                        sep_timetag=str(self._conf.sep_timetag),
                        dtype_sep=self.dtype_sep,
                        sep_index=str(sep_index)
                    )

                df = pd.read_csv(chunkname)

                logging.info("Appending %s to %s", chunkname, filename)
                if os.path.isfile(filename) is False:
                    df.to_csv(filename, index=False)
                else:
                    df.to_csv(filename, mode='a', header=False, index=False)

        if self._conf.data.test_clean_up:
            self.delete_concat_chunk_csv(num_chunks, models_dict, seg_id)

    def delete_concat_save_csv(self, models, items, output_name, sep_timetag):
        """Delete a series of csv files that concat_save_csv uses to concatenate and generate a single file. This method
        should only be called after concat_save_csv is being called.

        :param models: a dictionary of models that contains tag information
        :param items: a list containing the items (this is seg_list for the recommender engine)
        :param output_name: the prefix of the filenames (this is usually conf.workspace.name)
        :param sep_timetag: the sep_timetag for the model
        :return: None
        """
        logging.info("Deleting separate prediction files for segments")
        for tag, mdls in models.items():
            for seg_id in range(len(items)):
                try:
                    filename = PRED_PATH_SEG.format(
                        path=mdls.get(seg_id, mdls)["path"],
                        output_name=output_name,
                        tag=str(tag),
                        seg_id=str(seg_id),
                        sep_timetag=str(sep_timetag),
                        dtype_sep=self.dtype_sep
                    )

                    os.remove(filename)
                    logging.info("Deleting %s", filename)
                except FileNotFoundError as e:
                    logging.error(str(e))

    def delete_concat_chunk_csv(self, num_chunks, models_dict, seg_id=None):
        """Delete a series of csv files that were generated during the iteration through the prediction dataset. This
        method should only be called after concat_chunk_csv is being called.

        :param num_chunks: the number of chunks to be used for reading the file indices
        :param models_dict: an item in models_dict. The value for "path" is typically used in this class
        :param seg_id: segment id
        :return: None
        """
        logging.info("Deleting the files generated during iteration")

        for tag, mdls in models_dict.items():
            for index in range(num_chunks):
                sep_index = f"_{index}"
                if seg_id is None:
                    chunkname = PRED_PATH_CHUNK.format(
                        path=models_dict[tag]["path"],
                        output_name=self._conf.workspace.name,
                        tag=str(tag),
                        sep_timetag=str(self._conf.sep_timetag),
                        index=str(index),
                        dtype_sep=self.dtype_sep
                    )
                else:
                    mdl = mdls.get(seg_id, mdls)
                    chunkname = PRED_PATH_CHUNK_SEG.format(
                        path=mdl["path"],
                        output_name=self._conf.workspace.name,
                        tag=str(tag),
                        seg_id=str(seg_id),
                        sep_timetag=str(self._conf.sep_timetag),
                        dtype_sep=self.dtype_sep,
                        sep_index=sep_index,
                    )
                logging.info("Deleting %s", chunkname)
                os.remove(chunkname)

    def pickle_object(self, directory_path, filename, model_object, processor, feature_engineering):
        """Pickle the fitted model to training data and configuration object.

        :param directory_path: path the directory in which the model_object resides
        :param filename: name tag used for pickled object
        :param model_object: an object containing trained models, encoder, and scale objects
        :param processor: a DataProcessor object
        :param feature_engineering: an instance of FeatureEngineering class
        :return: None
        """

        logging.info("Pickling the model...")
        if not os.path.exists(directory_path):
            os.makedirs(directory_path)

        with open(filename, "wb") as pickle_file:
            pickle.dump((self.conf, model_object, processor, feature_engineering), pickle_file)


def pickle_object(conf, directory_path, filename, model_object, processor, feature_engineering):
    """Pickle the fitted model to training data and configuration object.

    :param conf: configuration file
    :param directory_path: path the directory in which the model_object resides
    :param filename: name tag used for pickled object
    :param model_object: an object containing trained models, encoder, and scale objects
    :param processor: a DataProcessor object
    :param feature_engineering: an instance of FeatureEngineering class
    :return: None
    """

    logging.info("Pickling the model...")
    if not os.path.exists(directory_path):
        os.makedirs(directory_path)

    with open(filename, "wb") as pickle_file:
        pickle.dump((conf, model_object, processor, feature_engineering), pickle_file)


def get_prediction_col_names(classes, prediction_type, column_name=None):
    """Take the number of classes for a model and generate a matching number of column names
    for prediction values to be appended to output data when writing predictions

    :param classes: a list of classes in the target
    :param prediction_type: type of prediction (proba, class, or all)
    :param column_name: output column name specified in the json file
    :return: a list of column names
    """
    # for binary classification return a single class
    if len(classes) == 2 or prediction_type == "class":
        return [column_name]
    else:
        return [column_name + "_" + str(c) for c in classes]



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module includes Learner's image classifiers. The pattern for image classifiers is no different than the
statndard classifiers. Here, we do not directly wrap sklearn models, etc. We write our own classes to wrap fastai's
implementations.
"""

import sys
import abc
import logging

import numpy as np
from torchvision.models.alexnet import alexnet
from torchvision.models.vgg import vgg11, vgg11_bn, vgg13, vgg13_bn, vgg16, vgg16_bn, vgg19, vgg19_bn
from torchvision.models.densenet import densenet121, densenet161, densenet169, densenet201
from torchvision.models.mobilenetv2 import mobilenet_v2
from torchvision.models.mobilenetv3 import mobilenet_v3_large, mobilenet_v3_small
from torchvision.models.mnasnet import mnasnet0_5, mnasnet1_0
from torchvision.models.resnet import resnet18, resnet34, resnet50, resnet101, resnet152, resnext50_32x4d, resnext101_32x8d, wide_resnet50_2, wide_resnet101_2
from torchvision.models.googlenet import googlenet
from torchvision.models.shufflenetv2 import shufflenet_v2_x0_5, shufflenet_v2_x1_0
import torch
import torch.nn as nn
from torch.nn.functional import softmax

from learner.model_manager.layer_manager import LayerBuilder
from learner.model_manager.loss_manager import LossBuilder
from learner.model_manager.optimizer_manager import OptimizerBuilder
from learner.model_manager.scheduler_manager import SchedulerBuilder, LRFinderScheduler
from learner.model_manager.nn_utils import freeze_params
from learner.model_manager.deep_loop_manager import DeepLoopManager
from learner.model_manager.nn_utils import ShapeFinder
from learner.callback_manager.callback import MetricLogger, ModelSaver, LRFinder


# these are the size of the last "features" layer after flattening them.
model_to_size = {"AlexNet": 9216,
                 "VGG11": 25088,
                 "VGG11_BN": 25088,
                 "VGG13": 25088,
                 "VGG13_BN": 25088,
                 "VGG16": 25088,
                 "VGG16_BN": 25088,
                 "VGG19": 25088,
                 "VGG19_BN": 25088,
                 "ResNet18": 512,
                 "ResNet34": 512,
                 "ResNet50": 2048,
                 "ResNet101": 2048,
                 "ResNet152": 2048,
                 "DenseNet121": 1024,
                 "DenseNet161": 2208,
                 "DenseNet169": 1664,
                 "DenseNet201": 1920,
                 "GoogLeNet": 1024,
                 "ShuffleNet_V2_x0_5": 1024,
                 "ShuffleNet_V2_x1_0": 1024,
                 "ShuffleNet_V2_x2_0": 1024,
                 "MobileNet_V2": 1280,
                 "MobileNet_V3_Large": 960,
                 "MobileNet_V3_Small": 576,
                 "ResNext50_32x4d": 2048,
                 "ResNext101_32x8d": 2048,
                 "Wide_ResNet50_2": 2048,
                 "Wide_ResNet101_2": 2048,
                 "MNASNet0_5": 1280,
                 "MNASNet1_0": 1280,
                 }

mdl_to_head = {"ImageClassifier": "classifier",
               "AlexNet": "classifier",
               "VGG11": "classifier",
               "VGG11_BN": "classifier",
               "VGG13": "classifier",
               "VGG13_BN": "classifier",
               "VGG16": "classifier",
               "VGG16_BN": "classifier",
               "VGG19": "classifier",
               "VGG19_BN": "classifier",
               "ResNet18": "fc",
               "ResNet34": "fc",
               "ResNet50": "fc",
               "ResNet101": "fc",
               "ResNet152": "fc",
               "DenseNet121": "classifier",
               "DenseNet161": "classifier",
               "DenseNet169": "classifier",
               "GoogLeNet": "fc",
               "ShuffleNet_V2_x0_5": "fc",
               "ShuffleNet_V2_x1_0": "fc",
               "DenseNet201": "classifier",
               "MobileNet_V2": "classifier",
               "MobileNet_V3_Large": "classifier",
               "MobileNet_V3_Small": "classifier",
               "ResNext50_32x4d": "fc",
               "ResNext101_32x8d": "fc",
               "Wide_ResNet50_2": "fc",
               "Wide_ResNet101_2": "fc",
               "MNASNet0_5": "classifier",
               "MNASNet1_0": "classifier",
               }


class AbstractImageClassifier:
    """The AbstractImageClassifier class. This class does not implement any functionality; it defines the structure of
    the child classes.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self):  # pragma: no cover
        self.mdl = None

    @staticmethod
    def learner_fit(tag, mdl, conf, train_loader, validation_loader):  # pragma: no cover
        pass

    @staticmethod
    def learner_predict(X, mdl, device, carry_data=True, full_data=None, cols_to_carry=None):  # pragma: no cover
        pass

    @staticmethod
    def learner_predict_proba(X, mdl, device, carry_data=True, full_data=None, cols_to_carry=None):  # pragma: no cover
        pass

    @staticmethod
    def learner_find_learning_rate(tag, mdl, conf, train_loader):
        pass


class ImageClassifierHandler(AbstractImageClassifier):
    """The handlers class calls the appropriate method of the image classifier classes with the appropriate arguments.
    """
    @abc.abstractmethod
    def __init__(self):
        super(AbstractImageClassifier, self).__init__()  # pragma: no cover

    @staticmethod
    def learner_fit(tag, mdl, conf, train_loader, validation_loader):
        try:
            return getattr(sys.modules[__name__], mdl["type"]).learner_fit(tag, mdl, conf, train_loader, validation_loader)
        except AttributeError:
            return StandardImageClassifier.learner_fit(tag, mdl, conf, train_loader, validation_loader)

    @staticmethod
    def learner_predict(test_data, mdl, device, carry_data=True, full_data=None, cols_to_carry=None):
        try:
            return getattr(sys.modules[__name__], mdl["type"]).learner_predict(test_data, mdl, device, carry_data, full_data, cols_to_carry)
        except AttributeError:
            return StandardImageClassifier.learner_predict(test_data, mdl, device, carry_data, full_data, cols_to_carry)

    @staticmethod
    def learner_predict_proba(test_data, mdl, device, carry_data=True, full_data=None, cols_to_carry=None):
        try:
            return getattr(sys.modules[__name__], mdl["type"]).learner_predict_proba(test_data, mdl, device, carry_data, full_data, cols_to_carry)
        except AttributeError:
            return StandardImageClassifier.learner_predict_proba(test_data, mdl, device, carry_data, full_data, cols_to_carry)

    @staticmethod
    def learner_find_learning_rate(tag, mdl, conf, train_loader):
        try:
            return getattr(sys.modules[__name__], mdl["type"]).learner_find_learning_rate(tag, mdl, conf, train_loader)
        except AttributeError:
            return StandardImageClassifier.learner_find_learning_rate(tag, mdl, conf, train_loader)


class StandardImageClassifier(AbstractImageClassifier):
    """A class for the standard methods of the image classifiers. These currently include learner_predict and
    learner_predict_proba methods. If a image classifier does not have these methods implemented, it would fall here.
    This way we won;t have to implement similar methods for all the image classifier models."""
    def __init__(self):
        super(AbstractImageClassifier, self).__init__()  # pragma: no cover

    @staticmethod
    def learner_fit(tag, mdl, conf, train_loader, validation_loader):
        """First, instantiate the model from PyTorch. If needed, freeze the parameters. Then replace the
        classifier attribute, build the loss function, and the optimizer. After that, use DeepTrainManager class to
        train and validate the model. In the end, return the trained model.

        :param tag: the model tag, this is an arbitrary string defined by the user.
        :param mdl: an item of the models_dict.
        :param conf: a conf object
        :param train_loader: the train_loader to loop through the training data.
        :param validation_loader: the validation_loader to loop through the validation data.
        :return: the trained model.
        """
        # instantiate the model
        model = getattr(sys.modules[__name__], mdl["type"].lower())(pretrained=True)
        # freeze the parameters if requested
        if mdl["params"]["freeze_features"]:
            model = freeze_params(model)

        layers = LayerBuilder(raw_list=mdl["params"]["classifier"], initial_size=model_to_size[mdl["type"]]).build_layers()

        setattr(model, mdl_to_head[mdl["type"]], nn.Sequential(*layers))
        params_to_update = list(getattr(model, mdl_to_head[mdl["type"]]).parameters())

        criterion = LossBuilder(mdl["params"]["loss"]).build_loss()
        optimizer = OptimizerBuilder(params_to_update if mdl["params"]["freeze_features"] else list(model.parameters()), mdl["params"]["optimizer"]).build_optimizer()
        scheduler = SchedulerBuilder(mdl["params"]["scheduler"] if "scheduler" in mdl["params"] else None).build_scheduler(optimizer)

        train_manager = DeepLoopManager(conf,
                                        model,
                                        criterion,
                                        optimizer,
                                        scheduler,
                                        mdl["params"]["epochs"],
                                        tag=tag,
                                        mdl=mdl,
                                        train_loader=train_loader,
                                        validation_loader=validation_loader,
                                        callbacks=[MetricLogger(), ModelSaver()])
        train_manager.train()
        return model

    @staticmethod
    def learner_find_learning_rate(tag, mdl, conf, train_loader):
        """First, instantiate the model from PyTorch. If needed, freeze the parameters. Then replace the
        classifier attribute, build the loss function, and the optimizer. Set the initial learning rate of the optimizer
        using the parameter defined in learning_rate_params of the configuration file. Also, compute the number of
        epochs using the number of iterations needed and the number of batches in the train_loader. After that,
        use DeepTrainManager class to find the learning rate. Here, we pass the LRFinder callback to the
        DeepTrainManager class. This callback handles things related to the learning rate parameters such as stopping
        the training loop, etc. In the end, return the trained model.

        :param tag: the model tag, this is an arbitrary string defined by the user.
        :param mdl: an item of the models_dict.
        :param conf: a conf object
        :param train_loader: the train_loader to loop through the training data.
        :return: the trained model.
        """
        # instantiate the model
        model = getattr(sys.modules[__name__], mdl["type"].lower())(pretrained=True)
        # freeze the parameters if requested
        if mdl["params"]["freeze_features"]:
            model = freeze_params(model)

        layers = LayerBuilder(raw_list=mdl["params"]["classifier"], initial_size=model_to_size[mdl["type"]]).build_layers()

        setattr(model, mdl_to_head[mdl["type"]], nn.Sequential(*layers))
        params_to_update = list(getattr(model, mdl_to_head[mdl["type"]]).parameters())

        criterion = LossBuilder(mdl["params"]["loss"]).build_loss()
        # set the initial lr to be start_lr irrespective of what's being set
        mdl["params"]["optimizer"]["lr"] = conf.model.lr_start_lr
        # determine how many epochs we need depending on the size of data and num_iter
        mdl["params"]["epochs"] = int(conf.model.lr_num_it / len(train_loader)) + 1

        optimizer = OptimizerBuilder(params_to_update if mdl["params"]["freeze_features"] else list(model.parameters()),
                                     mdl["params"]["optimizer"]).build_optimizer()

        scheduler = LRFinderScheduler(optimizer=optimizer,
                                      start_lr=conf.model.lr_start_lr,
                                      end_lr=conf.model.lr_end_lr,
                                      num_it=conf.model.lr_num_it)

        train_manager = DeepLoopManager(conf,
                                        model,
                                        criterion,
                                        optimizer,
                                        scheduler,
                                        mdl["params"]["epochs"],
                                        tag=tag,
                                        mdl=mdl,
                                        train_loader=train_loader,
                                        callbacks=[MetricLogger(),
                                                    LRFinder(conf=conf,
                                                             scheduler=scheduler,
                                                             stop_div=conf.model.lr_stop_div,
                                                             tag=tag,
                                                             mdl=mdl)])
        train_manager.train()
        return model

    @staticmethod
    def learner_predict(X, mdl, device, carry_data=True, full_data=None, cols_to_carry=None):
        """Use the input data and a mdl dictionary along with other arguments to get the class predictions. The value
        for the "model" key im mdl dictionary contains the trained model. This methods accepts the "device" argument
        as well. We move the data and the model to the appropriate device before making the predictions. This would
        enable making predictions using CPU or GPU.

        :param X: the input data for making the predictions.
        :param mdl: an item of the models_dict. This dictionary should contain the trained model.
        :param device: the device, this could be "cuda" or "cpu". The values are validated before they reach here.
        :param carry_data: a flag to indicate whether we should carry additional data or not.
        :param full_data: a pandas dataframe with additional data to send back if requested.
        :param cols_to_carry: The list of columns to select from the full dataset and send back if requested.
        :return: the predictions classes in numpy array as well as the selected columns from the data or None depending on the input.
        """
        with torch.no_grad():
            X = X.to(device)
            # put the model to evaluation mode. This is very important otherwise the results will not be correct
            mdl["model"].to(device).eval()
            pred = mdl["model"](X)
            # if the activation is LogSoftmax, do the exponential to get the probabilities
            if isinstance(getattr(mdl["model"], mdl_to_head[mdl["type"]])[-1], torch.nn.LogSoftmax):
                pred = torch.exp(pred)
            # if activation is not LogSoftmax and not Softmax, we need to pass it through a softmax function to get
            # the probabilities
            if not isinstance(getattr(mdl["model"], mdl_to_head[mdl["type"]])[-1], torch.nn.LogSoftmax) and \
               not isinstance(getattr(mdl["model"], mdl_to_head[mdl["type"]])[-1], torch.nn.Softmax):
                pred = torch.nn.functional.softmax(pred, dim=1)

            pred = np.argmax(pred.cpu(), 1)
            if carry_data:
                if full_data is not None and cols_to_carry:
                    return pred, full_data[cols_to_carry]
            return pred, None

    @staticmethod
    def learner_predict_proba(X, mdl, device, carry_data=True, full_data=None, cols_to_carry=None):
        """Use the input data and a mdl dictionary along with other arguments to get the predicted probabilities.
        The value for the "model" key im mdl dictionary contains the trained model. This methods accepts the "device"
        argument as well. We move the data and the model to the appropriate device before making the predictions.
        This would enable making predictions using CPU or GPU.

        :param X: the input data for making the predictions.
        :param mdl: an item of the models_dict. This dictionary should contain the trained model.
        :param device: the device, this could be "cuda" or "cpu". The values are validated before they reach here.
        :param carry_data: a flag to indicate whether we should carry additional data or not.
        :param full_data: a pandas dataframe with additional data to send back if requested.
        :param cols_to_carry: The list of columns to select from the full dataset and send back if requested.
        :return: the predictions classes in numpy array as well as the selected columns from the data or None depending on the input.
        """
        with torch.no_grad():
            X = X.to(device)
            # put the model to evaluation mode. This is very important otherwise the results will not be correct
            mdl["model"].to(device).eval()
            pred = mdl["model"](X).cpu()
            # if the activation is LogSoftmax, do the exponential to get the probabilities
            if isinstance(getattr(mdl["model"], mdl_to_head[mdl["type"]])[-1], torch.nn.LogSoftmax):
                pred = torch.exp(pred)
            # if activation is not LogSoftmax and not Softmax, we need to pass it through a softmax function to get
            # the probabilities
            if not isinstance(getattr(mdl["model"], mdl_to_head[mdl["type"]])[-1], torch.nn.LogSoftmax) and \
               not isinstance(getattr(mdl["model"], mdl_to_head[mdl["type"]])[-1], torch.nn.Softmax):
                pred = torch.nn.functional.softmax(pred, dim=1)

            if carry_data:
                if full_data is not None and cols_to_carry:
                    if len(mdl["classes"]) == 2:
                        return pred[:, 1], full_data[cols_to_carry]
                    else:
                        return pred, full_data[cols_to_carry]
            if len(mdl["classes"]) == 2:
                return pred[:, 1], None
            return pred, None


class ImageClassifier:
    """A class for building an image classifier model from the scratch. This class leverages LearnerImageClassifier
    to get the model object. The learner_fit is the main method here
    """

    @staticmethod
    def learner_fit(tag, mdl, conf, train_loader, validation_loader):
        """First, we need to get the dimensions of the images to pass them to the LearnerImageClassifier class.
        Since we are building the model from the scratch as opposed to transfer learning, we won't freeze
        the parameters. We then build the loss function and the optimizer. After that, we use
        DeepTrainManager class to train and validate the model. In the end, we return the trained model.

        :param tag: the model tag, this is an arbitrary string defined by the user.
        :param mdl: an item of the models_dict.
        :param conf: a conf object
        :param train_loader: the train_loader to loop through the training data.
        :param validation_loader: the validation_loader to loop through the validation data.
        :return: the trained model.
        """
        channel, height, width = ShapeFinder(train_loader, data_type="image").find_initial_shape()
        # instantiate the model
        model = LearnerImageClassifier(conf=conf, mdl=mdl, channel=channel, height=height, width=width)

        criterion = LossBuilder(mdl["params"]["loss"]).build_loss()
        optimizer = OptimizerBuilder(list(model.parameters()), mdl["params"]["optimizer"]).build_optimizer()
        scheduler = SchedulerBuilder(mdl["params"]["scheduler"] if "scheduler" in mdl["params"] else None).build_scheduler(optimizer)

        train_manager = DeepLoopManager(conf,
                                        model,
                                        criterion,
                                        optimizer,
                                        scheduler,
                                        mdl["params"]["epochs"],
                                        tag=tag,
                                        mdl=mdl,
                                        train_loader=train_loader,
                                        validation_loader=validation_loader,
                                        callbacks=[MetricLogger(), ModelSaver()])
        train_manager.train()
        return model

    @staticmethod
    def learner_find_learning_rate(tag, mdl, conf, train_loader):
        """First, we need to get the dimensions of the images to pass them to the LearnerImageClassifier class.
        Since we are building the model from the scratch as opposed to transfer learning, we won't freeze
        the parameters. We then build the loss function and the optimizer. We set the initial learning rate of the
        optimizer using the parameter defined in learning_rate_params of the configuration file. Also, compute the
        number of epochs using the number of iterations needed and the number of batches in the train_loader. After
        that, use DeepTrainManager class to find the learning rate. Here, we pass the LRFinder callback to the
        DeepTrainManager class. This callback handles things related to the learning rate parameters such as stopping
        the training loop, etc. In the end, return the trained model.

        :param tag: the model tag, this is an arbitrary string defined by the user.
        :param mdl: an item of the models_dict.
        :param conf: a conf object
        :param train_loader: the train_loader to loop through the training data.
        :return: the trained model.
        """
        channel, height, width = ShapeFinder(train_loader, data_type="image").find_initial_shape()
        # instantiate the model
        model = LearnerImageClassifier(conf=conf, mdl=mdl, channel=channel, height=height, width=width)

        criterion = LossBuilder(mdl["params"]["loss"]).build_loss()
        # set the initial lr to be start_lr irrespective of what's being set
        mdl["params"]["optimizer"]["lr"] = conf.model.lr_start_lr
        # determine how many epochs we need depending on the size of data and num_iter
        mdl["params"]["epochs"] = int(conf.model.lr_num_it / len(train_loader)) + 1

        optimizer = OptimizerBuilder(list(model.parameters()),
                                     mdl["params"]["optimizer"]).build_optimizer()

        scheduler = LRFinderScheduler(optimizer=optimizer,
                                      start_lr=conf.model.lr_start_lr,
                                      end_lr=conf.model.lr_end_lr,
                                      num_it=conf.model.lr_num_it)

        train_manager = DeepLoopManager(conf,
                                        model,
                                        criterion,
                                        optimizer,
                                        scheduler,
                                        mdl["params"]["epochs"],
                                        tag=tag,
                                        mdl=mdl,
                                        train_loader=train_loader,
                                        callbacks=[MetricLogger(),
                                                    LRFinder(conf=conf,
                                                             scheduler=scheduler,
                                                             stop_div=conf.model.lr_stop_div,
                                                             tag=tag,
                                                             mdl=mdl)])
        train_manager.train()
        return model


class LearnerImageClassifier(nn.Module):
    """The class to implement an image classifier model from scratch using the architecture defined by the user. The
    image classifier model has two major attributes namely the "features" and the "classifier"."""
    def __init__(self, conf, mdl, channel, height, width):
        """Initialize a LearnerImageClassifier object using the input arguments.

        :param conf: a conf object.
        :param mdl: an item of models_dict.
        :param channel: the number of channels of the input image.
        :param height: the height of the image.
        :param width: the width of the image.
        """
        super(LearnerImageClassifier, self).__init__()
        self._conf = conf
        self._mdl = mdl
        self._channel = channel
        self._height = height
        self._width = width
        layer_builder = LayerBuilder(raw_list=self._mdl["params"]["features"],
                                     initial_size=self._channel,
                                     height=self._height,
                                     width=self._width)
        features = layer_builder.build_layers()
        self.features = nn.Sequential(*features)

        classifier = LayerBuilder(raw_list=self._mdl["params"]["classifier"],
                                  initial_size=layer_builder.running_flatten_size).build_layers()

        self.classifier = nn.Sequential(*classifier)

    def forward(self, x):
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        return x



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

# monkey-patch smtplib so we don't send actual emails
smtp=None
inbox=[]


class Message(object):
    def __init__(self, from_address, to_address, fullmessage):
        self.from_address = from_address
        self.to_address = to_address
        self.fullmessage = fullmessage


class DummySMTP(object):
    def __init__(self, address, port):
        self.address = address
        self.port = port
        self.username = None
        self.password = None
        self.has_quit = False
        global smtp
        smtp = self

    def login(self, username, password):
        self.username = username
        self.password = password

    def starttls(self):
        return None

    def sendmail(self, from_address, to_address, fullmessage):
        global inbox
        inbox.append(Message(from_address, to_address, fullmessage))
        return []

    def quit(self):
        self.has_quit = True


# this is the actual monkey patch (simply replacing one class with another)
import smtplib
smtplib.SMTP = DummySMTP



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module implements the base classes for deep engines. We typically do not instantiate this base classes. We have
moved the base deep class into a separate class to make things a bit more clear"""

import logging
import os
import sys

import torch
import pandas as pd

from learner.analysis.analysis import Analysis
from learner.configuration.configuration import Configuration
from learner.data_worker.deep_tabular_data_manager import TrainDataManager, ValidationDataManager, TestDataManager
from learner.data_worker.output_handler import OutputHandler
from learner.model_manager.model_initializer import ModelInitializer, get_pickle_dir
from learner.model_manager.deep_classifiers import DeepClassifierHandler
from learner.model_manager.deep_regressors import DeepRegressorHandler
from learner.model_manager.prediction_manager import DeepClassifierPredictor, DeepRegressorPredictor
from learner.utilities.timer import timeit
from learner.validator.output_validator import OutputValidator


class BaseDeep(ModelInitializer, TrainDataManager):
    """The base class for tabular deep learning engines including DeepClassifier and DeepRegressor. This class implement
    most of the methods for these engines because they are highly similar. Please note that engine_manager does not
    instantiate this class.
    """
    def __init__(self, conf: Configuration):
        """Instantiate an object of the BaseDeep class using an instance of the Configuration class

        :param conf: an object of the Configuration class
        """
        super(BaseDeep, self).__init__(conf)
        super(ModelInitializer, self).__init__(conf)

        self.conf.model.nrows_score = 0
        # this will hold information about all models as opposed to a single model. This is used to combine the
        # predictions of multiple models
        self.loaded_models_dict = {}

    @timeit("build the model(s)")
    def run_engine(self):
        """The main method that communicates with other methods to drive the engine.

        :return: None
        """
        if self._conf.model.train_models:
            self.train_models()
            self.predict()
        else:
            logging.info("Loading saved model...")
            self.load_models(self._conf.workspace.path)

        self.validate_output()

    @timeit("train the model(s)")
    def train_models(self):
        """Loop through all the items in models_dict and call the train_models method to train each model.

        :return: None
        """
        logging.info("Building the models...")
        for tag, mdl in self.models_dict.items():
            logging.info(f"Fitting {tag} model")
            # train the model and save it into model dictionary
            # depending on the user input, either do grid_search, cross_validation, or a single fit
            self.fit_models(tag, mdl)

    def fit_models(self, tag, mdl):
        """Call appropriate methods to find the learning rate or fit the model. There are a lot of similarities
        between this method and the method in for image_classifier engines. We need to maintain that consistency.

        :param tag: the tag of the model (arbitrary tag defined by user)
        :param mdl: an items in models_dict
        :return: None
        """
        train_data_manager = TrainDataManager(self.conf, mdl)
        train_loader = train_data_manager.get_data_loader()
        validation_loader = None

        # we need this assignment because we'll need it if we want to make predictions right away
        self.processor = train_data_manager.processor
        self.feature_engineering = train_data_manager.feature_engineering
        if self._conf.data.validation_score_activate:
            validation_data_manager = ValidationDataManager(self.conf,
                                                            mdl,
                                                            processor=self.processor,
                                                            feature_engineering=self.feature_engineering)
            validation_loader = validation_data_manager.get_data_loader()

        if self._conf.model.lr_activate:
            logging.info("Finding the learning rate...")
            mdl["model"] = getattr(sys.modules[__name__], self._conf.engine + "Handler").\
                learner_find_learning_rate(tag=tag,
                                           mdl=mdl,
                                           conf=self.conf,
                                           train_loader=train_loader)
        else:
            mdl["model"] = getattr(sys.modules[__name__], self._conf.engine + "Handler").\
                learner_fit(tag=tag,
                            mdl=mdl,
                            conf=self.conf,
                            train_loader=train_loader,
                            validation_loader=validation_loader)

    @timeit("make predictions using the trained model(s)")
    def predict(self):
        """Call predict_test if we need to make predictions on test data. Scoring the validation data happens as we
        train the models.

        :return: None
        """
        if self._conf.data.test_prediction_activate:
            logging.info("Making predictions using test data...")
            self.predict_test()

    @timeit("make predictions using the trained model(s)")
    def predict_test(self):
        """Load the test data in chunks and make predictions. This method instantiates the predictor class of the
        engine and calls make_predictions method on them. To make sure, we can make predictions on the dataset of
        any sizes, we implement a nested loop here. First, we get a data reader object. We then load a chunk of data
        and pass it to the get_data_loader to obtain a data_loader object. A confusing thing here could be saving the
        output. As you can see, we use the output handler for segmentation to save the predictions and do the
        concatenation. This is because, similar to segmentation, we are using nested loops and dealing with two
        dimensional output matrices namely chunks and batches. We can think of a chunk index like segment index, and
        the batch index like the chunk index in the segmentation. We may need to update the underlying output handler
        classes in the future to remove some of the confusions.

        :return: None
        """
        test_data_manager = TestDataManager(self._conf, self.processor, self.feature_engineering)
        reader = test_data_manager.get_reader()

        num_chunks = 0
        output = OutputHandler(self._conf, data_type="test")
        for reader_index, chunk in enumerate(reader):
            chunk = test_data_manager.get_reader_data(chunk)
            num_chunks += 1
            for tag, mdl in self.models_dict.items():
                data_loader = test_data_manager.get_data_loader(chunk, mdl)
                for loader_index, (X, data) in enumerate(data_loader):
                    # data will be a dictionary ready to be converted to a dataframe
                    data = pd.DataFrame(data=data)
                    predictor = getattr(sys.modules[__name__], self._conf.engine + "Predictor")(mdl, self._conf)
                    prediction_df, data = predictor.make_predictions(X,
                                                                     data,
                                                                     data_type="test",
                                                                     column_name=self._conf.data.test_column_name,
                                                                     prediction_type=self._conf.data.test_prediction_type)

                    predictor.write_predictions(tag,
                                                prediction_df,
                                                data,
                                                seg_id=reader_index,
                                                index=loader_index,
                                                mdl=mdl,
                                                data_type="test",
                                                add_timetag=self.conf.data.test_add_timetag,
                                                add_version=self.conf.model.version_activate,
                                                version_name=self.conf.model.version_version_name,
                                                version_column_name=self.conf.model.version_column_name
                                                )
                # we stitch the chunks from dataloader together
                output.concat_chunk_csv(len(data_loader), {tag: mdl}, seg_id=reader_index)
            # keep track of number of rows in test dataset for validation purposes
            self.conf.model.nrows_score += chunk.shape[0]

        output.concat_chunk_csv(num_chunks, self.models_dict)

    @timeit("load the model(s)")
    def load_models(self, output_path):
        """Load saved models from the provided path if full directory is provided. Otherwise, find and load the latest
        saved models. Here, we also make the call to the combine_predictions method.

        :param output_path: directory path to the saved models
        :return: None
        """
        _, pickled_file_names, _ = get_pickle_dir(output_path, ext="pth")
        # this is to prevent IDEs from complaining
        sep_timetag = ""
        for file_name in pickled_file_names:
            checkpoint = torch.load(file_name)
            self.models_dict = {checkpoint["tag"]: checkpoint["mdl"]}
            self.loaded_models_dict[checkpoint["tag"]] = self.models_dict[checkpoint["tag"]]
            checkpoint["conf"].sep_timetag = self.conf.sep_timetag
            # we want to use the current data parameters
            checkpoint["conf"].data = self._conf.data
            # we want to use the current combine parameters
            checkpoint["conf"].combine = self._conf.combine
            self.processor = checkpoint["processor"]
            self.feature_engineering = checkpoint["feature_engineering"]
            self._conf = checkpoint["conf"]
            self.predict()

    @timeit("validate the output")
    def validate_output(self):
        """Validate the predictions and the files created by the engine. Currently, we only validate the number of
        predictions to make sure it's consistent with the original test data.

        :return: None
        """
        if self._conf.validation.validate and self._conf.data.test_prediction_activate:
            output_validator = OutputValidator(self.conf, self.models_dict)
            output_validator.validate()



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""Learner uses a callback/event handling mechanism to modify the train loops of deep learning models. This module
implements the main classes. The two classes "Event" and "EventManager" will always be sitting in this module but other
classes that handle different events may move to other modules."""
import math
import logging
from shutil import copyfile

import pandas as pd
import torch

from learner.utilities.templates import SAVE_MODEL_PATH, SAVE_MODEL_PATH_EPOCH, LEARNING_RATE
from learner.analysis.plot import Plot


class Callback:
    """The base class for all event handler classes. All of the event handler classes inherit from this class. This
    class has one method for all the events. This way, the child classes would only need to implement the relevant
    methods and the rest will be handled by the methods in the parent class."""
    def on_batch_end(self, *args, **kwargs):
        pass

    def on_epoch_end(self, *args, **kwargs):
        pass


class CallbackManager:
    """The EventManager class accepts a list of callback objects. This class implements all the methods of the Event
    class. In each method, this class loops through each object and calls the instance method identical to the caller
    method. For example, the method "on_epoch_end" would loop through all callback objects and would call the
    "on_epoch_end" method of those objects/instances"""
    def __init__(self, *callbacks):
        """Initialize an EventManager object using a list of callback instances.

        :param callbacks: a list of callback objects/instances
        """
        self.callbacks = callbacks

    def on_batch_end(self, *args, **kwargs):
        """Loop through the list of callbacks and call their on_epoch_end method.

        :param args: a tuple/list of positional arguments
        :param kwargs: a dictionary of keyword arguments
        :return: None
        """
        for callback in self.callbacks:
            callback.on_batch_end(*args, **kwargs)

    def on_epoch_end(self, *args, **kwargs):
        """Loop through the list of callbacks and call their on_epoch_end method.

        :param args: a tuple/list of positional arguments
        :param kwargs: a dictionary of keyword arguments
        :return: None
        """
        for callback in self.callbacks:
            callback.on_epoch_end(*args, **kwargs)


class MetricLogger(Callback):
    """A class to log the metrics and other information during the training and validation process."""
    def on_epoch_end(self, *args, **kwargs):
        """This method expects the keyword argument "metrics", which is a dictionary containing the metrics and their
        corresponding values. The metrics can be anything. Here, we iterate through each items to create a logging
        message. We then log that message. The items in the "metrics" dictionary would depend on the user input and
        other factors.

        :param args: a tuple/list of positional arguments
        :param kwargs: a dictionary of keyword arguments
        :return: None
        """
        message = ""
        for metric, value in kwargs["metrics"].items():
            if isinstance(value, int):
                message += f"{metric}: {value}\t"
            else:
                message += f"{metric}: {value:.10f}\t"
        logging.info(message)


class ModelSaver(Callback):
    """A class to save the trained model and other necessary information during the training process. The user can decide
    the frequency in which the models should be saved. Once those conditions are met, Learner saved the models."""
    def on_epoch_end(self, *args, **kwargs):
        """At the end of each epoch, we need to decide whether we should save the model or not. The user can define a
        parameter called "save_interval" to define the frequency of saving the models. This method needs a lot of
        keyword arguments such as "mdl", "conf", "tag", etc. The reason is that, Learner needs to be able to load the
        saved model and resume the training or make the predictions with minimum input from the user.

        :param args: a tuple/list of positional arguments
        :param kwargs: a dictionary of keyword arguments
        :return: None
        """
        if kwargs["mdl"]["params"]["save_interval"] and kwargs["epoch"] % kwargs["mdl"]["params"]["save_interval"] == 0:
            logging.info("Saving the model...")
            filename = SAVE_MODEL_PATH.format(
                path=kwargs["conf"].model.models_dict[kwargs["tag"]]["path"],
                output_name=kwargs["conf"].workspace.name,
                tag=kwargs["tag"],
                sep_timetag=str(kwargs["conf"].sep_timetag),
                ext=".pth"
            )
            epoch_filename = SAVE_MODEL_PATH_EPOCH.format(
                path=kwargs["conf"].model.models_dict[kwargs["tag"]]["path"],
                output_name=kwargs["conf"].workspace.name,
                tag=kwargs["tag"],
                epoch=kwargs["epoch"],
                sep_timetag=str(kwargs["conf"].sep_timetag),
                ext=".bck"
            )
            kwargs["mdl"]["model"] = kwargs["model"]
            torch.save({"conf": kwargs["conf"],
                        "tag": kwargs["tag"],
                        "mdl": kwargs["mdl"],
                        "processor": kwargs["processor"],
                        "feature_engineering": kwargs["feature_engineering"]},
                       filename)
            copyfile(filename, epoch_filename)


class LRFinder(Callback):
    """This callback class is used for finding the optimum learning rate. This class is only used when we want to find
    the learning rate. See this article for more details: https://sgugger.github.io/how-do-you-find-a-good-learning-rate.html"""
    def __init__(self, conf, scheduler, stop_div, tag, mdl):
        """Initializer a LRFinder object using the input arguments.

        :param conf: a conf object
        :param scheduler: a scheduler object. The scheduler must have the step method.
        :param stop_div: a flag to specify if we should stop when the loss starts to diverge.
        :param tag: the model tag, this is an arbitrary string defined by the user.
        :param mdl: an item of the models_dict.
        """
        self._conf = conf
        self._scheduler = scheduler
        self.best_loss = torch.tensor(0)
        self.stop_div = stop_div
        self.tag = tag
        self.mdl = mdl
        self.avg_loss = 0
        self.beta = 0.98
        self.data = {}

    @property
    def conf(self):
        return self._conf

    @property
    def scheduler(self):
        return self._scheduler

    def on_batch_end(self, *args, **kwargs):
        """At the end of each batch (not epoch), we'd need to compute the average loss, the smoothed loss, record the
        smoothed loss, and update the learning rate. If we are done or if the loss is diverging, we'd plot the loss as a
        function of learning rate and tell the trainer to stop.

        :param args: a tuple/list of positional arguments
        :param kwargs: a dictionary of keyword arguments
        :return: None
        """
        #  compute the smoothed loss
        self.avg_loss = self.beta * self.avg_loss + (1-self.beta) * kwargs["loss"].item()
        smoothed_loss = self.avg_loss / (1 - self.beta**(kwargs["iteration"] + 1))

        if kwargs["iteration"] == 0 or smoothed_loss < self.best_loss:
            self.best_loss = kwargs["loss"]
        self.record_data(smoothed_loss)
        self._scheduler.step()
        if self._scheduler.is_done or (self.stop_div and (smoothed_loss > 10 * self.best_loss or math.isnan(smoothed_loss))):
            df = self.get_df()
            self.plot_lr(df)
            kwargs["deep_trainer"].should_stop = True

    def record_data(self, smoothed_loss):
        """Update the data dictionary with adding the current learning rate and the smoothed loss.

        :param smoothed_loss: the value of the smoothed loss
        :return: None
        """
        self.data[self._scheduler.optimizer.param_groups[0]["lr"]] = smoothed_loss

    def get_df(self):
        """Convert the data dictionary to a pandas dataframe to make the plotting easier.

        :return: a pandas dataframe containing two columns: "lr" and "loss"
        """
        return pd.DataFrame(data=self.data.items(), columns=["lr", "loss"])

    def plot_lr(self, df):
        """Use the dataframe containing the learning rate and the loss and create a plot.

        :param df: a pandas dataframe containing two columns: "lr" and "loss"
        :return: None
        """
        plot_filename = LEARNING_RATE.format(
            path=self._conf.model.models_dict[self.tag]["path"],
            output_name=self._conf.workspace.name,
            tag=str(self.tag),
            sep_timetag=str(self._conf.sep_timetag))

        plot = Plot(self._conf)
        plot.learning_rate(df["lr"],
                           df["loss"],
                           filename=plot_filename)
